/*
 * jWinkeyServer, a cwdeamon replacement that interfaces to Winkeyer by K1EL.
 * Copyright (C) 2012  Andy Summers, G4KNO (g4kno.mail@gmail.com).
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */

package winkeyserver;

import java.awt.event.KeyEvent;
import java.net.*;
import java.util.ArrayList;
import java.util.Observable;

/**
 * Class UdpServer.
 * Creates a cwdaemon-like UDP server and reports received packets to an Observer.
 */
public class UdpServer extends Observable implements Runnable {
    
    private boolean serverSocketActive = false;
    private DatagramSocket serverSocket;
    private ArrayList rxData = new ArrayList();
    
    // Constructor.
    public UdpServer() {     
    }
    
    // Method to start server on specified port.
    public void startServer(int portNo) {
        try {
            serverSocketActive = true;
            serverSocket = new DatagramSocket(portNo);
        } catch (Exception e){
            serverSocketActive = false;
            System.err.println(e);
        }
    } // startServer.
    
    // Method to stop server.
    public void stopServer() {
        serverSocketActive = false;
        serverSocket.close();
    } // stopServer.
    

    @Override
    // Concrete implementation of Runnable's run() method.
    public void run() {
        
        while (serverSocketActive) {
        
            try {
                // Create a new buffer each time to make sure it's empty.
                byte[] buf = new byte[1024];
                DatagramPacket rxPacket = new DatagramPacket(buf, buf.length);
                serverSocket.receive(rxPacket);
                String data = new String(rxPacket.getData());
                InetAddress IPAddress = rxPacket.getAddress();
                //System.out.println(IPAddress.getHostAddress() + ": " + data);
                
                // Clear the ArrayList for the new rxPacket.
                rxData.clear();
                
                // cwdaemon identifies control codes using a <ESC> prefix.
                if ( data.charAt(0) == ((char) KeyEvent.VK_ESCAPE) ) {
                    // It's a control code.
                    
                    switch (data.charAt(1)) {
                        
                        case '0':
                            // Reset to default values.
                            // Unsupported - use this application's defaults.
                            break;
                            
                        case '2':
                            // Speed value.
                            rxData.add("speed");
                            rxData.add(IPAddress.getHostAddress());
                            rxData.add(data.substring(2).trim());
                            break;
                            
                        case '3':
                            // Tone value (300 ~ 1000Hz),(0 turns sound off).
                            rxData.add("tone");
                            rxData.add(IPAddress.getHostAddress());
                            rxData.add(data.substring(2).trim());
                            break;
                            
                        case '4':
                            // Abort message.
                            rxData.add("abort");
                            rxData.add(IPAddress.getHostAddress());
                            break;
                            
                        case '5':
                            // Stop (exit) server.
                            // Unsupported - control server only from this application.
                            break;
                            
                        case '6':
                            // Set uninterruptable mode.
                            // What's this??
                            break;
                        
                        case '7':
                            // Weight value (-50 ~ 50).
                            rxData.add("weight");
                            rxData.add(IPAddress.getHostAddress());
                            rxData.add(data.substring(2).trim());
                            break;
                            
                        case '8':
                            // Set device for keying.
                            // Unsupported - makes no sense in this application.
                            break;
                            
                        case '9':
                            // Port No. (obsolete).
                            // Unsupported.
                            break;
                            
                        case 'a':
                            // PTT keying.
                            rxData.add("ptt");
                            rxData.add(IPAddress.getHostAddress());
                            rxData.add(data.substring(2).trim());
                            break;
                            
                        case 'b':
                            // Mic/soundcard switching.
                            // Unsupported - makes no sense in this application.
                            break;
                            
                        case 'c':
                            // Tune x seconds long (cwdaemon's max: 10s)
                            // (winkey's max: 99s)
                            rxData.add("tune");
                            rxData.add(IPAddress.getHostAddress());
                            rxData.add(data.substring(2).trim());
                            break;
                            
                        case 'd':
                            // PTT on delay (0 ~ 50ms).
                            // Unspupported - use winkey GUI.
                            break;
                            
                        case 'e':
                            // Band index.
                            // Unsupported - makes no sense in this application.
                            break;
                            
                        case 'f':
                            // Sound system.
                            // Partially supported - accept 'n' (none) & 'c' console buzzer.
                            rxData.add("sound");
                            rxData.add(IPAddress.getHostAddress());
                            if (data.charAt(2) == 'c') {
                                rxData.add("1");
                            } else if (data.charAt(2) == 'n') {
                                rxData.add("0");
                            }
                            break;
                            
                        case 'g':
                            // Volume.
                            // Unsupported - winkey can't vary volume.
                            break;
                            
                        case 'h':
                            // Cwdaemon debug mode.
                            // Unsupported.
                            break;
                            
                        default:
                            // Invalid control key.
                            System.err.println("Received invalid control key");
                    }
                    
                    
                    
                } else {
                    // It's a cw message.
                    rxData.add("message");
                    rxData.add(IPAddress.getHostAddress());
                    // We MUST remove the newline char at the end for Winkey.
                    rxData.add(data.trim());
                    
                }
                                
                // Fire observer event.
                setChanged();
                notifyObservers(rxData);
                
            } catch (Exception e) {
                System.err.println(e);
            }
        
        }
        
    } // run.
    
} // UdpServer.
