/*  
 * jWinkeyServer, a cwdeamon replacement that interfaces to Winkeyer by K1EL.
 * Copyright (C) 2012  Andy Summers, G4KNO (g4kno.mail@gmail.com).
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */

package winkeyserver;

import java.io.*;
import java.util.*;
import purejavacomm.*;

/**
 * Class WinkeySerialComms.
 * Handles the serial communications to Winkey.
 */
public class WinkeySerialComms extends Observable implements Runnable {
    
    // Class variables.
    private SerialPort      serialPort;
    private InputStream     in;
    private OutputStream    out;
    private ArrayList       rxData = new ArrayList();
    private int             winkeyVersion;
    
    // Status bits.
    protected boolean echoTestMode = false;
    protected boolean xoff = false;
    protected boolean breakin = false;
    protected boolean busy = false;
    protected boolean keydown = false;
    protected boolean wait = false;
    
    
    // Mask constants for determining the type of data received.
    static final int statusMask = 0xC0;
    static final int speedPotMask = 0x80;
    
    
    // Constructor.
    public WinkeySerialComms() {
        
    }
    
    
    @Override
    // Concrete implementation of Runnable's run() method.
    public void run() {
        // Not used.
        // It may not be necessary to create a new thread for WinkeySerialComms,
        // but it allows synchronized methods to be used, so I don't have to
        // worry about contention between the GUI and the udpServer.
    }
    
    
    
    // Method to list the available comm ports.
    public ArrayList getAvailableSerialPorts() {
        
        ArrayList portsArrayList = new ArrayList();
        
        //System.out.println("getting available serial ports");
        
        CommPortIdentifier portid = null;
	Enumeration e = CommPortIdentifier.getPortIdentifiers();
	while (e.hasMoreElements()) {
            portid = (CommPortIdentifier) e.nextElement();
            if (portid.getPortType() == CommPortIdentifier.PORT_SERIAL) {
                //System.out.println("found " + portid.getName());
                portsArrayList.add(portid.getName());
            }
	}
        return portsArrayList;
        
    } // getAvailableSerialPorts.
    
    
    // Method to open a port by name.
    public void openSerialPort(String name) {
        
        CommPortIdentifier portId = null;
        
        try {
            portId = CommPortIdentifier.getPortIdentifier(name);
            serialPort = (SerialPort) portId.open("SerialComms", 2000);
            serialPort.setSerialPortParams( 1200,
                                            SerialPort.DATABITS_8,
                                            SerialPort.STOPBITS_2,
                                            SerialPort.PARITY_NONE);
            
            in = serialPort.getInputStream();
            out = serialPort.getOutputStream();
                
            serialPort.addEventListener(new SerialReader(in));
            serialPort.notifyOnDataAvailable(true);
            
        } catch (Exception e) {
            System.err.println(e.toString());
        }
        
    } // openSerialPort.
    
    // Method to close winkey comms.
    public synchronized void closeWinkey() {
        // Issue host close followed by serial port close.
        if (!(serialPort == null)) {
            hostClose();
            serialPort.close();
        }
    } // closeWinkey.
    
    
    // Method to initiate host open mode.
    public synchronized void hostOpen() {
        // Reinitialise the version variable (hostOpen responds with version)
        winkeyVersion = 0;
        // Admin(Host Open).
        byte b[] = {0x00, 0x02};
        sendBytes(b);
    } // hostOpen.
    
    
    // Method to initiate host close mode.
    public synchronized void hostClose() {
        // Admin(Host Close).
        byte b[] = {0x00, 0x03};
        sendBytes(b);
    } // hostClose.
    
    
    // Method to load defaults into winkey hardware.
    public synchronized void loadDefaults(byte mode, byte wpmSpeed, byte sidetone,
                            byte weight, byte pttLead, byte pttTail, byte wpmMin,
                            byte wpmRange, byte extn1, byte keyComp, byte wpmFarns,
                            byte paddleSwitchPoint, byte ditDahRatio,
                            byte pinConfig) {
        byte b[] = {0x0F, mode, wpmSpeed, sidetone, weight, pttLead, pttTail,
                    wpmMin, wpmRange, extn1, keyComp, wpmFarns,
                    paddleSwitchPoint, ditDahRatio, pinConfig, 0x00};
        sendBytes(b);
    } // loadDefaults.
    
    
    // Method to send text.
    public synchronized void sendText(String text) {
        sendBytes(new String(text.toUpperCase()).getBytes());
    } // sendText.
    
    
    // Method to set WPM speed.
    public synchronized void setWpmSpeed(byte speed) {
        byte b[] = {0x02, speed};
        sendBytes(b);
    } // setWpmSpeed.
    
    // Method to get WPM speed (return value caught by SerialReader)
    public synchronized void getWpmSpeed() {
        byte b[] = {0x07};
        sendBytes(b);
    } // getWpmSpeed.
    
    
    // Method to setup speed pot.
    public synchronized void setUpSpeedPot(byte minSpeed, byte speedRange) {
        byte b[] = {0x05, minSpeed, speedRange, 0x00};
        sendBytes(b);
    } // setUpSpeedPot.
    
    
    // Method to set Dah/Dit ratio.
    public synchronized void setDahDitRatio(byte ratio) {
        byte b[] = {0x17, ratio};
        sendBytes(b);
    } // setDahDitRatio.
    
    
    // Method to set weighting.
    public synchronized void setWeighting(byte weight) {
        byte b[] = {0x03, weight};
        sendBytes(b);
    } // setWeighting.
    
    
    // Method to set Farnsworth WPM.
    public synchronized void setFarnsWpm(byte wpm) {
        byte b[] = {0x0D, wpm};
        sendBytes(b);
    } // setFarnsWpm.
    
    
    // Method to set Winkeyer mode.
    public synchronized void setWinkeyerMode(byte mode) {
        byte b[] = {0x0E, mode};
        sendBytes(b);
    } // setWinkeyerMode.
    
    
    // Method to set Winkeyer pin config.
    public synchronized void setPinConfig(byte config) {
        byte b[] = {0x09, config};
        sendBytes(b);
    } // setPinConfig.
    
    
    // Method to set Winkeyer paddle switchpoint.
    public synchronized void setPaddleSwitchpoint(byte switchpoint) {
        byte b[] = {0x12, switchpoint};
        sendBytes(b);
    } // setPaddleSwitchpoint.
    
    
    // Method to set Winkeyer sidetone.
    public synchronized void setSidetone(byte sidetone) {
        byte b[] = {0x01, sidetone};
        sendBytes(b);
    } // setSidetone.
    
    
    // Method to set Winkeyer PTT lead/tail.
    public synchronized void setPttLeadTail(byte lead, byte tail) {
        byte b[] = {0x04, lead, tail};
        sendBytes(b);
    } // setPttLeadTail.
    
    
    // Method to set Winkeyer key compensation.
    public synchronized void setKeyComp(byte comp) {
        byte b[] = {0x11, comp};
        sendBytes(b);
    } // setKeyComp.
    
    
    // Method to set Winkeyer 1st extension.
    public synchronized void setFirstExtn(byte extn) {
        byte b[] = {0x10, extn};
        sendBytes(b);
    } // setFirstExtn.
    
    
    // Method to send a NOP to Winkeyer.
    public synchronized void sendNop() {
        byte b[] = {0x13};
        sendBytes(b);
    } // sendNop.
    
    
    // Method to set Winkeyer Echo Test Mode.
    public synchronized void setEchoTest(byte echoChar) {
        echoTestMode = true;
        byte b[] = {0x00, 0x04, echoChar};
        sendBytes(b);
    } // setEchoTest.
    
    
    // Method to clear winkey buffer (aborts sending).
    public synchronized void clearBuffer() {
        byte b[] = {0x0A};
        sendBytes(b);
    } // clearBuffer.
    
    
    // Method to key PTT for custom purpose (only effective if sidetone & PTT are disabled.
    public synchronized void setPtt(byte pttState) {
        byte b[] = {0x18, pttState};
        sendBytes(b);
    } // setPtt.
    
    
    // Method to set the key output. An alternative method to setPtt() that avoids
    // the necessity to disable winkey PTT.
    public synchronized void keyImmediate(byte keyState) {
        byte b[] = {0x0B, keyState};
        sendBytes(b);
    } // keyImmediate.
    
    
    // Method to assert key output for keyDuration seconds.
    public synchronized void keyBuffered(byte keyDuration) {
        byte b[] = {0x19, keyDuration};
        sendBytes(b);
    } // keyBuffered.
    
    
    // Method to place a speed change command into the buffer.
    public synchronized void bufferedSpeedChange(byte newSpeed) {
        byte b[] = {0x1C, newSpeed};
        sendBytes(b);
    } // bufferedSpeedChange.
    
    
    // Method to cancel a speed change command (reverts to original setting).
    public synchronized void cancelBufferedSpeedChange() {
        byte b[] = {0x1E};
        sendBytes(b);
    } // cancelBufferedSpeedChange.
    
    
    // Utility method to send bytes to Winkeyer.
    public synchronized void sendBytes(byte[] bytes) {
        try {
            out.write(bytes);
        } catch (Exception e) {
            System.err.println(e.toString());
        }
    } // sendBytes.

    
    // Anonymous inner class to handle data coming from the serial port.
    public class SerialReader implements SerialPortEventListener {
        private InputStream in;
        
        public SerialReader ( InputStream in ) {
            this.in = in;
        }
        
        public void serialEvent(SerialPortEvent arg0) {
            
            int data;
            
            try {
                data = in.read();
                //System.out.println("RXD DATA = 0x" + Integer.toHexString(data));
                
                if ( (data & statusMask) == statusMask ) {
                    // I've only coded WK1 compatible mode,
                    // so pushbutton isn't supported.
                    
                    rxData.clear();
                    rxData.add("status");
                    
                    if ((data & 0x01) == 0x01) {
                        rxData.add("xoff");
                        xoff = true;
                    } else {
                        xoff = false;
                    }
                    if ((data & 0x02) == 0x02) {
                        rxData.add("breakin");
                        breakin = true;
                    } else {
                        breakin = false;
                    }
                    if ((data & 0x04) == 0x04) {
                        rxData.add("busy");
                        busy = true;
                    } else {
                        busy = false;
                    }
                    if ((data & 0x08) == 0x08) {
                        rxData.add("keydown");
                        keydown = true;
                    } else {
                        keydown = false;
                    }
                    if ((data & 0x10) == 0x10) {
                        rxData.add("wait");
                        wait = true;
                    } else {
                        wait = false;
                    }
                    if ((data & 0x1F) == 0x00) {
                        rxData.add("null");
                    }
                } else if ( (data & statusMask) == speedPotMask ) {
                    // Speed pot byte.
                    // The host may or may not have requested this.
                    rxData.clear();
                    rxData.add("speedPot");
                    rxData.add(Integer.toHexString(data - speedPotMask));
                } else {
                    
                    if (breakin) {
                        rxData.clear();
                        rxData.add("paddleEcho");
                        rxData.add(Integer.toHexString(data));
                    } else if (echoTestMode) {
                        rxData.clear();
                        rxData.add("serialEcho");
                        rxData.add(Integer.toHexString(data));
                        echoTestMode = false;
                    } else {
                        // The only other thing it could be is the version No.
                        winkeyVersion = data;
                        rxData.clear();
                        rxData.add("version");
                        rxData.add(Integer.toHexString(data));
                    }
                
                }
                    
                setChanged();
                notifyObservers(rxData);
                
            } catch (Exception e) {
                System.err.println(e.toString());
            }
            
            
        } // serialEvent.

    } // SerialReader.


} // SerialComms.
