/*
 * jWinkeyServer, a cwdeamon replacement that interfaces to Winkeyer by K1EL.
 * Copyright (C) 2012  Andy Summers, G4KNO (g4kno.mail@gmail.com).
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */

package winkeyserver;

import java.awt.event.KeyEvent;
import java.util.*;
import javax.swing.DefaultComboBoxModel;
import javax.swing.text.DefaultCaret;


/**
 * Class WinkeyServerGUI.
 * A GUI to setup & observe messages sent to Winkeyer via a cwdaemon-like UDP
 * server.
 */
public class WinkeyServerGUI extends javax.swing.JFrame {
    
    private UdpServer udpServer;
    private WinkeySerialComms wsc;
    private Settings s = Settings.getInstance();
    private DefaultComboBoxModel cbm;
    
    // A useful variable used to prevent initialisation of the form
    // from firing events.
    private boolean initialising = false;
    
    // We must have a local variable for the WPM sent to winkey because it
    // uses '0' to tell it to use the pot setting.
    private String WINKEY_WPM = null;
    
    private boolean versionPending = false;
    private boolean serialEchoPending = false;
    public int version = 0;
    public int serialEchoByte = 0;

    /**
     * Creates new form WinkeyServerGUI
     */
    public WinkeyServerGUI() {
        
        initialising = true;
        initComponents();
        
        // Position the frame on display.
        this.setLocation(Integer.parseInt(s.POSITION_X),
                        Integer.parseInt(s.POSITION_Y));
        
        // Start UDP server in a new thread.
        udpServer = new UdpServer();
        udpServer.addObserver(new UdpServerObserver());
        udpServer.startServer(Integer.parseInt(s.CWDAEMON_PORT));      
        new Thread(udpServer).start();
        
        
        // Start serial comms.
        wsc = new WinkeySerialComms();
        wsc.addObserver(new WinkeyObserver());
        new Thread(wsc).start();
        
        //Add the available ports to the combo box.
        cbm = new DefaultComboBoxModel(wsc.getAvailableSerialPorts().toArray());
        serialPortComboBox.setModel(cbm);
        
        // Must initialise combo box after setting model or it uses first in list.
        serialPortComboBox.setSelectedItem(s.SERIAL_PORT);
        
        //*****************************************************
        //Initialise the rest of the form with the values already stored.
        //*****************************************************
        cwdaemonPortTextField.setText(s.CWDAEMON_PORT);
        // Note we need to display in decimal, but store in hex.
        hostWpmTextField.setText(Integer.toString(Integer.parseInt(s.HOST_WPM, 16)));
        minWpmTextField.setText(Integer.toString(Integer.parseInt(s.WINKEY_MIN_WPM, 16)));
        maxWpmTextField.setText(Integer.toString(
                Integer.parseInt(s.WINKEY_MIN_WPM, 16)
                + Integer.parseInt(s.WINKEY_WPM_RANGE, 16)
                ));
        
        if (s.HOST_SETS_SPEED.equalsIgnoreCase("true")) {
            hostSetsSpeedButton.setSelected(true);
            hostWpmTextField.setEnabled(true);
            minWpmTextField.setEnabled(false);
            maxWpmTextField.setEnabled(false);
            potWpmTextField.setEnabled(false);
            WINKEY_WPM = s.HOST_WPM;
            
        } else {
            potSetsSpeedButton.setSelected(true);
            hostWpmTextField.setEnabled(false);
            minWpmTextField.setEnabled(true);
            maxWpmTextField.setEnabled(true);
            potWpmTextField.setEnabled(true);
            WINKEY_WPM = "00";                      // Tell winkey to use pot.
        }
        
        int sidetoneControl = Integer.parseInt(s.WINKEY_SIDETONE, 16);
        if ((sidetoneControl & 0x80) == 0x80) {
            paddleOnlySidetoneCheckBox.setSelected(true);
        } else {
            paddleOnlySidetoneCheckBox.setSelected(false);
        }
        sideToneFreqComboBox.setSelectedIndex(10 - sidetoneControl & 0x0F);
        
        weightingSpinner.setValue(Integer.parseInt(s.WINKEY_WEIGHT, 16));
        
        pttLeadSpinner.setValue(Integer.parseInt(s.WINKEY_PTT_LEAD, 16));
        pttTailSpinner.setValue(Integer.parseInt(s.WINKEY_PTT_TAIL, 16));
        
        int pinConfig = Integer.parseInt(s.WINKEY_PIN_CONFIG, 16);
        if ((pinConfig & 0x01) == 0x01) {
            pttEnableCheckBox.setSelected(true);
        } else {
            pttEnableCheckBox.setSelected(false);
        }
        if ((pinConfig & 0x02) == 0x02) {
            sidetoneEnableCheckBox.setSelected(true);
        } else {
            sidetoneEnableCheckBox.setSelected(false);
        }
        if ((pinConfig & 0x04) == 0x04) {
            keyOut2CheckBox.setSelected(true);
        } else {
            keyOut2CheckBox.setSelected(false);
        }
        if ((pinConfig & 0x08) == 0x08) {
            keyOut1CheckBox.setSelected(true);
        } else {
            keyOut1CheckBox.setSelected(false);
        }
        
        hangTimeComboBox.setSelectedIndex((pinConfig & 0x30) >> 4);
        
        int priority = (pinConfig & 0xC0) >> 6;    // Use later with Mode.
        
        farnsWpmSpinner.setValue(Integer.parseInt(s.WINKEY_WPM_FARNS, 16));
        
        int mode = Integer.parseInt(s.WINKEY_MODE, 16);
        if ((mode & 0x01) == 0x01) {
            contestSpacingCheckBox.setSelected(true);
        } else {
            contestSpacingCheckBox.setSelected(false);
        }
        if ((mode & 0x02) == 0x02) {
            autospaceCheckBox.setSelected(true);
        } else {
            autospaceCheckBox.setSelected(false);
        }
        if ((mode & 0x08) == 0x08) {
            paddleSwapCheckBox.setSelected(true);
        } else {
            paddleSwapCheckBox.setSelected(false);
        }
        if ((mode & 0x80) == 0x80) {
            paddleWatchDogDisableCheckBox.setSelected(true);
        } else {
            paddleWatchDogDisableCheckBox.setSelected(false);
        }
        
        int keyMode = (mode & 0x30) >> 4;
        if (keyMode < 0x2) {
            keyModeComboBox.setSelectedIndex(keyMode);
        } else if (keyMode == 0x2) {
            keyModeComboBox.setSelectedIndex(keyMode + priority);
        } else {
            keyModeComboBox.setSelectedIndex(keyMode + 2);
        }
        
        firstExtnSpinner.setValue(Integer.parseInt(s.WINKEY_EXTN1, 16));
        keyCompSpinner.setValue(Integer.parseInt(s.WINKEY_KEY_COMP, 16));
        paddleSwitchpointSpinner.setValue(Integer.parseInt(s.WINKEY_PADDLE_SWITCHPOINT, 16));
        
        float dahDitRatio = (float) Integer.parseInt(s.WINKEY_DAH_DIT_RATIO, 16);
        dahDitRatioSpinner.setValue(3*dahDitRatio/50);
        
        if (s.CWDAEMON_CONTROLS_PTT.equalsIgnoreCase("true")) {
            controlsPttRadioButton.setSelected(true);
        } else {
            controlsKeyRadioButton.setSelected(true);
        }
        
        // Set the debugTextArea to automatically scroll down.
        DefaultCaret caret = (DefaultCaret) debugTextArea.getCaret();
        caret.setUpdatePolicy(DefaultCaret.ALWAYS_UPDATE);
        
        // Set the maximum number of text lines in the debugTextArea.
        debugTextArea.getDocument().addDocumentListener(
            new LimitLinesDocumentListener(Integer.parseInt(s.MAX_TEXTLINES)));
        
        // Credits.
        debugTextArea.append("*********************************************\n");
        debugTextArea.append("jWinkeyServer version 0.1\n");
        debugTextArea.append("Copyright (C) 2012, Andy Summers, G4KNO.\n");
        debugTextArea.append("This software comes with\nABSOLUTELY NO WARRANTY.\n");
        debugTextArea.append("*********************************************\n");
        
        // Now try to start winkey.
        restartWinkey();
        
        initialising = false;
        
    }
    
    // Method to start or restart the connection to winkey.
    public void restartWinkey() {
        
        debugTextArea.append("\n");
        
        wsc.openSerialPort(s.SERIAL_PORT);
        
        // Wait 400ms to allow Winkey to init from power-up.
        try {
            Thread.sleep(400);
        } catch (Exception e) {
            System.err.println(e);
        }
        
        //Send four NOPs.
        for (int i=0; i<4; i++) {
            wsc.sendNop();
        }
        
        // Do an echo test on the serial interface.
        int echoTestData = 0x3F;
        wsc.setEchoTest((byte) echoTestData);
        // Since rxd data is in a separate thread, we need to wait up to 2s for it.
        long endTime = System.currentTimeMillis() + 2000;
        while (!serialEchoPending) {
            // It seems you have to do something in the while loop here, or the
            // serialEchoPending variable isn't updated properly (bug!?). So do some
            // sleeping to reduce CPU load.
            try {
                Thread.sleep(10);
            } catch (Exception e) {
                System.err.println(e);
            }
            if (System.currentTimeMillis() >= endTime) {
                debugTextArea.append("Timed-out waiting for serial echo byte!\n");
                debugTextArea.append("Check serial interface!\n");
                break;
            }
        }
        if (serialEchoPending && (serialEchoByte == echoTestData)) {
            
            serialEchoPending = false;
        
            wsc.hostOpen();
            // Winkey should reply to hostOpen with version No.
            // Since rxd data is in a separate thread, we need to wait up to 2s for it.
            endTime = System.currentTimeMillis() + 2000;
            while (!versionPending) {
                // It seems you have to do something in the while loop here, or the
                // versionPending variable isn't updated properly (bug!?). So do some
                // sleeping to reduce CPU load.
                try {
                    Thread.sleep(10);
                } catch (Exception e) {
                    System.err.println(e);
                }
                if (System.currentTimeMillis() >= endTime) {
                    debugTextArea.append("Timed-out waiting for Winkey version number!\n");
                    break;
                }
            }
            if (versionPending) {

                versionPending = false;

                wsc.loadDefaults((byte) Integer.parseInt(s.WINKEY_MODE, 16),
                                (byte) Integer.parseInt(WINKEY_WPM, 16),
                                (byte) Integer.parseInt(s.WINKEY_SIDETONE, 16),
                                (byte) Integer.parseInt(s.WINKEY_WEIGHT, 16),
                                (byte) Integer.parseInt(s.WINKEY_PTT_LEAD, 16),
                                (byte) Integer.parseInt(s.WINKEY_PTT_TAIL, 16),
                                (byte) Integer.parseInt(s.WINKEY_MIN_WPM, 16),
                                (byte) Integer.parseInt(s.WINKEY_WPM_RANGE, 16),
                                (byte) Integer.parseInt(s.WINKEY_EXTN1, 16),
                                (byte) Integer.parseInt(s.WINKEY_KEY_COMP, 16),
                                (byte) Integer.parseInt(s.WINKEY_WPM_FARNS, 16),
                                (byte) Integer.parseInt(s.WINKEY_PADDLE_SWITCHPOINT, 16),
                                (byte) Integer.parseInt(s.WINKEY_DAH_DIT_RATIO, 16),
                                (byte) Integer.parseInt(s.WINKEY_PIN_CONFIG, 16));
                // Now get the current pot setting.
                wsc.getWpmSpeed();
                debugTextArea.append("Winkeyer ready!\n\n");
            }
        } else {
            debugTextArea.append("Incorrect serial echo byte!\n");
            debugTextArea.append("Check serial interface!\n");
        }
        
    } // restartWinkey.
    
    // Utility method to gracefully close the application.
    public void close() {
        udpServer.stopServer();
        wsc.closeWinkey();
    } // close

    /**
     * This method is called from within the constructor to initialize the form.
     * WARNING: Do NOT modify this code. The content of this method is always
     * regenerated by the Form Editor.
     */
    @SuppressWarnings("unchecked")
    // <editor-fold defaultstate="collapsed" desc="Generated Code">//GEN-BEGIN:initComponents
    private void initComponents() {

        speedButtonGroup = new javax.swing.ButtonGroup();
        pttButtonGroup = new javax.swing.ButtonGroup();
        jTabbedPane1 = new javax.swing.JTabbedPane();
        basicPanel = new javax.swing.JPanel();
        jScrollPane1 = new javax.swing.JScrollPane();
        debugTextArea = new javax.swing.JTextArea();
        cwdaemonPortTextField = new javax.swing.JTextField();
        serialPortComboBox = new javax.swing.JComboBox();
        jLabel2 = new javax.swing.JLabel();
        jLabel1 = new javax.swing.JLabel();
        restartWinkeyer = new javax.swing.JButton();
        jPanel1 = new javax.swing.JPanel();
        hostSetsSpeedButton = new javax.swing.JRadioButton();
        jLabel3 = new javax.swing.JLabel();
        hostWpmTextField = new javax.swing.JTextField();
        potSetsSpeedButton = new javax.swing.JRadioButton();
        jLabel4 = new javax.swing.JLabel();
        jLabel5 = new javax.swing.JLabel();
        jLabel6 = new javax.swing.JLabel();
        minWpmTextField = new javax.swing.JTextField();
        maxWpmTextField = new javax.swing.JTextField();
        potWpmTextField = new javax.swing.JTextField();
        cwControlPanel = new javax.swing.JPanel();
        sideToneFreqComboBox = new javax.swing.JComboBox();
        jLabel7 = new javax.swing.JLabel();
        jLabel8 = new javax.swing.JLabel();
        paddleOnlySidetoneCheckBox = new javax.swing.JCheckBox();
        sidetoneEnableCheckBox = new javax.swing.JCheckBox();
        jLabel12 = new javax.swing.JLabel();
        farnsWpmSpinner = new javax.swing.JSpinner();
        weightingSpinner = new javax.swing.JSpinner();
        paddleWatchDogDisableCheckBox = new javax.swing.JCheckBox();
        keyModeComboBox = new javax.swing.JComboBox();
        jLabel13 = new javax.swing.JLabel();
        paddleSwapCheckBox = new javax.swing.JCheckBox();
        autospaceCheckBox = new javax.swing.JCheckBox();
        contestSpacingCheckBox = new javax.swing.JCheckBox();
        paddleSwitchpointSpinner = new javax.swing.JSpinner();
        jLabel16 = new javax.swing.JLabel();
        jLabel17 = new javax.swing.JLabel();
        dahDitRatioSpinner = new javax.swing.JSpinner();
        txControlPanel = new javax.swing.JPanel();
        pttTailSpinner = new javax.swing.JSpinner();
        pttLeadSpinner = new javax.swing.JSpinner();
        jLabel9 = new javax.swing.JLabel();
        jLabel10 = new javax.swing.JLabel();
        pttEnableCheckBox = new javax.swing.JCheckBox();
        keyOut1CheckBox = new javax.swing.JCheckBox();
        keyOut2CheckBox = new javax.swing.JCheckBox();
        hangTimeComboBox = new javax.swing.JComboBox();
        jLabel11 = new javax.swing.JLabel();
        jLabel15 = new javax.swing.JLabel();
        keyCompSpinner = new javax.swing.JSpinner();
        jLabel14 = new javax.swing.JLabel();
        firstExtnSpinner = new javax.swing.JSpinner();
        jPanel2 = new javax.swing.JPanel();
        controlsPttRadioButton = new javax.swing.JRadioButton();
        controlsKeyRadioButton = new javax.swing.JRadioButton();

        setDefaultCloseOperation(javax.swing.WindowConstants.EXIT_ON_CLOSE);
        setTitle("jWinkeyServer");
        setPreferredSize(new java.awt.Dimension(350, 500));
        setResizable(false);
        addComponentListener(new java.awt.event.ComponentAdapter() {
            public void componentMoved(java.awt.event.ComponentEvent evt) {
                formComponentMoved(evt);
            }
        });

        basicPanel.setPreferredSize(new java.awt.Dimension(500, 567));

        debugTextArea.setColumns(20);
        debugTextArea.setRows(5);
        debugTextArea.addKeyListener(new java.awt.event.KeyAdapter() {
            public void keyPressed(java.awt.event.KeyEvent evt) {
                debugTextAreaKeyPressed(evt);
            }
        });
        jScrollPane1.setViewportView(debugTextArea);

        cwdaemonPortTextField.addKeyListener(new java.awt.event.KeyAdapter() {
            public void keyReleased(java.awt.event.KeyEvent evt) {
                cwdaemonPortTextFieldKeyReleased(evt);
            }
        });

        serialPortComboBox.setEditable(true);
        serialPortComboBox.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                serialPortComboBoxActionPerformed(evt);
            }
        });

        jLabel2.setText("cwdaemon Port");

        jLabel1.setText("Serial Port");

        restartWinkeyer.setText("restart winkeyer");
        restartWinkeyer.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                restartWinkeyerActionPerformed(evt);
            }
        });

        jPanel1.setBorder(javax.swing.BorderFactory.createLineBorder(new java.awt.Color(0, 0, 0)));

        speedButtonGroup.add(hostSetsSpeedButton);
        hostSetsSpeedButton.setText("Host sets speed");
        hostSetsSpeedButton.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                hostSetsSpeedButtonActionPerformed(evt);
            }
        });

        jLabel3.setText("Host WPM");

        hostWpmTextField.addKeyListener(new java.awt.event.KeyAdapter() {
            public void keyReleased(java.awt.event.KeyEvent evt) {
                hostWpmTextFieldKeyReleased(evt);
            }
        });

        speedButtonGroup.add(potSetsSpeedButton);
        potSetsSpeedButton.setText("Pot sets speed");
        potSetsSpeedButton.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                potSetsSpeedButtonActionPerformed(evt);
            }
        });

        jLabel4.setText("Min WPM");

        jLabel5.setText("Max WPM");

        jLabel6.setText("Pot WPM");

        minWpmTextField.addKeyListener(new java.awt.event.KeyAdapter() {
            public void keyReleased(java.awt.event.KeyEvent evt) {
                minWpmTextFieldKeyReleased(evt);
            }
        });

        maxWpmTextField.addKeyListener(new java.awt.event.KeyAdapter() {
            public void keyReleased(java.awt.event.KeyEvent evt) {
                maxWpmTextFieldKeyReleased(evt);
            }
        });

        javax.swing.GroupLayout jPanel1Layout = new javax.swing.GroupLayout(jPanel1);
        jPanel1.setLayout(jPanel1Layout);
        jPanel1Layout.setHorizontalGroup(
            jPanel1Layout.createParallelGroup(javax.swing.GroupLayout.Alignment.LEADING)
            .addGroup(javax.swing.GroupLayout.Alignment.TRAILING, jPanel1Layout.createSequentialGroup()
                .addContainerGap()
                .addGroup(jPanel1Layout.createParallelGroup(javax.swing.GroupLayout.Alignment.LEADING, false)
                    .addComponent(hostSetsSpeedButton)
                    .addGroup(jPanel1Layout.createSequentialGroup()
                        .addComponent(jLabel3)
                        .addPreferredGap(javax.swing.LayoutStyle.ComponentPlacement.RELATED, javax.swing.GroupLayout.DEFAULT_SIZE, Short.MAX_VALUE)
                        .addComponent(hostWpmTextField, javax.swing.GroupLayout.PREFERRED_SIZE, 50, javax.swing.GroupLayout.PREFERRED_SIZE)))
                .addGap(18, 18, 18)
                .addGroup(jPanel1Layout.createParallelGroup(javax.swing.GroupLayout.Alignment.LEADING)
                    .addGroup(jPanel1Layout.createSequentialGroup()
                        .addGroup(jPanel1Layout.createParallelGroup(javax.swing.GroupLayout.Alignment.LEADING)
                            .addComponent(jLabel5)
                            .addComponent(jLabel6)
                            .addComponent(jLabel4))
                        .addGap(18, 18, 18)
                        .addGroup(jPanel1Layout.createParallelGroup(javax.swing.GroupLayout.Alignment.LEADING)
                            .addComponent(maxWpmTextField, javax.swing.GroupLayout.PREFERRED_SIZE, 50, javax.swing.GroupLayout.PREFERRED_SIZE)
                            .addComponent(potWpmTextField, javax.swing.GroupLayout.PREFERRED_SIZE, 50, javax.swing.GroupLayout.PREFERRED_SIZE)
                            .addComponent(minWpmTextField, javax.swing.GroupLayout.PREFERRED_SIZE, 50, javax.swing.GroupLayout.PREFERRED_SIZE)))
                    .addComponent(potSetsSpeedButton))
                .addContainerGap(javax.swing.GroupLayout.DEFAULT_SIZE, Short.MAX_VALUE))
        );
        jPanel1Layout.setVerticalGroup(
            jPanel1Layout.createParallelGroup(javax.swing.GroupLayout.Alignment.LEADING)
            .addGroup(jPanel1Layout.createSequentialGroup()
                .addContainerGap()
                .addGroup(jPanel1Layout.createParallelGroup(javax.swing.GroupLayout.Alignment.LEADING)
                    .addComponent(hostSetsSpeedButton)
                    .addComponent(potSetsSpeedButton, javax.swing.GroupLayout.Alignment.TRAILING))
                .addPreferredGap(javax.swing.LayoutStyle.ComponentPlacement.RELATED)
                .addGroup(jPanel1Layout.createParallelGroup(javax.swing.GroupLayout.Alignment.BASELINE)
                    .addComponent(jLabel3)
                    .addComponent(hostWpmTextField, javax.swing.GroupLayout.PREFERRED_SIZE, javax.swing.GroupLayout.DEFAULT_SIZE, javax.swing.GroupLayout.PREFERRED_SIZE)
                    .addComponent(jLabel4)
                    .addComponent(minWpmTextField, javax.swing.GroupLayout.PREFERRED_SIZE, javax.swing.GroupLayout.DEFAULT_SIZE, javax.swing.GroupLayout.PREFERRED_SIZE))
                .addPreferredGap(javax.swing.LayoutStyle.ComponentPlacement.RELATED)
                .addGroup(jPanel1Layout.createParallelGroup(javax.swing.GroupLayout.Alignment.BASELINE)
                    .addComponent(maxWpmTextField, javax.swing.GroupLayout.PREFERRED_SIZE, javax.swing.GroupLayout.DEFAULT_SIZE, javax.swing.GroupLayout.PREFERRED_SIZE)
                    .addComponent(jLabel5))
                .addPreferredGap(javax.swing.LayoutStyle.ComponentPlacement.RELATED)
                .addGroup(jPanel1Layout.createParallelGroup(javax.swing.GroupLayout.Alignment.BASELINE)
                    .addComponent(potWpmTextField, javax.swing.GroupLayout.PREFERRED_SIZE, javax.swing.GroupLayout.DEFAULT_SIZE, javax.swing.GroupLayout.PREFERRED_SIZE)
                    .addComponent(jLabel6))
                .addContainerGap(javax.swing.GroupLayout.DEFAULT_SIZE, Short.MAX_VALUE))
        );

        javax.swing.GroupLayout basicPanelLayout = new javax.swing.GroupLayout(basicPanel);
        basicPanel.setLayout(basicPanelLayout);
        basicPanelLayout.setHorizontalGroup(
            basicPanelLayout.createParallelGroup(javax.swing.GroupLayout.Alignment.LEADING)
            .addGroup(basicPanelLayout.createSequentialGroup()
                .addContainerGap()
                .addGroup(basicPanelLayout.createParallelGroup(javax.swing.GroupLayout.Alignment.LEADING, false)
                    .addComponent(jScrollPane1)
                    .addGroup(basicPanelLayout.createSequentialGroup()
                        .addGroup(basicPanelLayout.createParallelGroup(javax.swing.GroupLayout.Alignment.TRAILING, false)
                            .addGroup(basicPanelLayout.createSequentialGroup()
                                .addComponent(jLabel2)
                                .addPreferredGap(javax.swing.LayoutStyle.ComponentPlacement.RELATED)
                                .addComponent(cwdaemonPortTextField, javax.swing.GroupLayout.PREFERRED_SIZE, 91, javax.swing.GroupLayout.PREFERRED_SIZE))
                            .addGroup(javax.swing.GroupLayout.Alignment.LEADING, basicPanelLayout.createSequentialGroup()
                                .addComponent(jLabel1)
                                .addGap(18, 18, 18)
                                .addComponent(serialPortComboBox, javax.swing.GroupLayout.PREFERRED_SIZE, 117, javax.swing.GroupLayout.PREFERRED_SIZE)))
                        .addPreferredGap(javax.swing.LayoutStyle.ComponentPlacement.RELATED)
                        .addComponent(restartWinkeyer))
                    .addComponent(jPanel1, javax.swing.GroupLayout.DEFAULT_SIZE, javax.swing.GroupLayout.DEFAULT_SIZE, Short.MAX_VALUE))
                .addContainerGap(javax.swing.GroupLayout.DEFAULT_SIZE, Short.MAX_VALUE))
        );
        basicPanelLayout.setVerticalGroup(
            basicPanelLayout.createParallelGroup(javax.swing.GroupLayout.Alignment.LEADING)
            .addGroup(basicPanelLayout.createSequentialGroup()
                .addGap(18, 18, 18)
                .addGroup(basicPanelLayout.createParallelGroup(javax.swing.GroupLayout.Alignment.BASELINE)
                    .addComponent(jLabel2)
                    .addComponent(cwdaemonPortTextField, javax.swing.GroupLayout.PREFERRED_SIZE, javax.swing.GroupLayout.DEFAULT_SIZE, javax.swing.GroupLayout.PREFERRED_SIZE))
                .addPreferredGap(javax.swing.LayoutStyle.ComponentPlacement.RELATED)
                .addGroup(basicPanelLayout.createParallelGroup(javax.swing.GroupLayout.Alignment.BASELINE)
                    .addComponent(serialPortComboBox, javax.swing.GroupLayout.PREFERRED_SIZE, javax.swing.GroupLayout.DEFAULT_SIZE, javax.swing.GroupLayout.PREFERRED_SIZE)
                    .addComponent(jLabel1)
                    .addComponent(restartWinkeyer))
                .addPreferredGap(javax.swing.LayoutStyle.ComponentPlacement.RELATED)
                .addComponent(jPanel1, javax.swing.GroupLayout.PREFERRED_SIZE, javax.swing.GroupLayout.DEFAULT_SIZE, javax.swing.GroupLayout.PREFERRED_SIZE)
                .addPreferredGap(javax.swing.LayoutStyle.ComponentPlacement.RELATED)
                .addComponent(jScrollPane1, javax.swing.GroupLayout.DEFAULT_SIZE, 179, Short.MAX_VALUE)
                .addContainerGap())
        );

        jTabbedPane1.addTab("Basic", basicPanel);

        sideToneFreqComboBox.setModel(new javax.swing.DefaultComboBoxModel(new String[] { "400", "444", "500", "571", "666", "800", "1000", "1333", "2000", "4000" }));
        sideToneFreqComboBox.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                sideToneFreqComboBoxActionPerformed(evt);
            }
        });

        jLabel7.setText("Sidetone Freq [Hz]");

        jLabel8.setText("Weighting [%]");

        paddleOnlySidetoneCheckBox.setText("Paddle Only Sidetone");
        paddleOnlySidetoneCheckBox.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                paddleOnlySidetoneCheckBoxActionPerformed(evt);
            }
        });

        sidetoneEnableCheckBox.setText("Sidetone Enable");
        sidetoneEnableCheckBox.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                sidetoneEnableCheckBoxActionPerformed(evt);
            }
        });

        jLabel12.setText("Farnsworth WPM");

        farnsWpmSpinner.setModel(new javax.swing.SpinnerNumberModel(18, 10, 99, 1));
        farnsWpmSpinner.addChangeListener(new javax.swing.event.ChangeListener() {
            public void stateChanged(javax.swing.event.ChangeEvent evt) {
                farnsWpmSpinnerStateChanged(evt);
            }
        });

        weightingSpinner.setModel(new javax.swing.SpinnerNumberModel(50, 10, 90, 1));
        weightingSpinner.addChangeListener(new javax.swing.event.ChangeListener() {
            public void stateChanged(javax.swing.event.ChangeEvent evt) {
                weightingSpinnerStateChanged(evt);
            }
        });

        paddleWatchDogDisableCheckBox.setText("Paddle Watchdog Disable");
        paddleWatchDogDisableCheckBox.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                paddleWatchDogDisableCheckBoxActionPerformed(evt);
            }
        });

        keyModeComboBox.setModel(new javax.swing.DefaultComboBoxModel(new String[] { "Iambic B", "Iambic A", "Ultimatic", "Ultimatic (dahs)", "Ultimatic (dits)", "Bug Mode" }));
        keyModeComboBox.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                keyModeComboBoxActionPerformed(evt);
            }
        });

        jLabel13.setText("Key Mode");

        paddleSwapCheckBox.setText("Paddle Swap");
        paddleSwapCheckBox.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                paddleSwapCheckBoxActionPerformed(evt);
            }
        });

        autospaceCheckBox.setText("Autospace");
        autospaceCheckBox.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                autospaceCheckBoxActionPerformed(evt);
            }
        });

        contestSpacingCheckBox.setText("Contest Spacing");
        contestSpacingCheckBox.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                contestSpacingCheckBoxActionPerformed(evt);
            }
        });

        paddleSwitchpointSpinner.setModel(new javax.swing.SpinnerNumberModel(50, 10, 90, 1));
        paddleSwitchpointSpinner.addChangeListener(new javax.swing.event.ChangeListener() {
            public void stateChanged(javax.swing.event.ChangeEvent evt) {
                paddleSwitchpointSpinnerStateChanged(evt);
            }
        });

        jLabel16.setText("Paddle Switchpoint [%]");

        jLabel17.setText("Dah/Dit Ratio");

        dahDitRatioSpinner.setModel(new javax.swing.SpinnerNumberModel(Float.valueOf(3.0f), Float.valueOf(1.98f), Float.valueOf(3.96f), Float.valueOf(0.06f)));
        dahDitRatioSpinner.addChangeListener(new javax.swing.event.ChangeListener() {
            public void stateChanged(javax.swing.event.ChangeEvent evt) {
                dahDitRatioSpinnerStateChanged(evt);
            }
        });

        javax.swing.GroupLayout cwControlPanelLayout = new javax.swing.GroupLayout(cwControlPanel);
        cwControlPanel.setLayout(cwControlPanelLayout);
        cwControlPanelLayout.setHorizontalGroup(
            cwControlPanelLayout.createParallelGroup(javax.swing.GroupLayout.Alignment.LEADING)
            .addGroup(cwControlPanelLayout.createSequentialGroup()
                .addContainerGap()
                .addGroup(cwControlPanelLayout.createParallelGroup(javax.swing.GroupLayout.Alignment.LEADING)
                    .addGroup(cwControlPanelLayout.createSequentialGroup()
                        .addGroup(cwControlPanelLayout.createParallelGroup(javax.swing.GroupLayout.Alignment.LEADING)
                            .addComponent(jLabel8)
                            .addComponent(jLabel12)
                            .addComponent(jLabel13))
                        .addGap(45, 45, 45)
                        .addGroup(cwControlPanelLayout.createParallelGroup(javax.swing.GroupLayout.Alignment.LEADING)
                            .addComponent(farnsWpmSpinner, javax.swing.GroupLayout.PREFERRED_SIZE, javax.swing.GroupLayout.DEFAULT_SIZE, javax.swing.GroupLayout.PREFERRED_SIZE)
                            .addComponent(keyModeComboBox, javax.swing.GroupLayout.PREFERRED_SIZE, javax.swing.GroupLayout.DEFAULT_SIZE, javax.swing.GroupLayout.PREFERRED_SIZE)
                            .addComponent(dahDitRatioSpinner, javax.swing.GroupLayout.PREFERRED_SIZE, javax.swing.GroupLayout.DEFAULT_SIZE, javax.swing.GroupLayout.PREFERRED_SIZE)
                            .addComponent(weightingSpinner, javax.swing.GroupLayout.PREFERRED_SIZE, javax.swing.GroupLayout.DEFAULT_SIZE, javax.swing.GroupLayout.PREFERRED_SIZE)))
                    .addComponent(jLabel17)
                    .addGroup(cwControlPanelLayout.createSequentialGroup()
                        .addComponent(jLabel16)
                        .addPreferredGap(javax.swing.LayoutStyle.ComponentPlacement.RELATED)
                        .addComponent(paddleSwitchpointSpinner, javax.swing.GroupLayout.PREFERRED_SIZE, javax.swing.GroupLayout.DEFAULT_SIZE, javax.swing.GroupLayout.PREFERRED_SIZE))
                    .addGroup(cwControlPanelLayout.createSequentialGroup()
                        .addGroup(cwControlPanelLayout.createParallelGroup(javax.swing.GroupLayout.Alignment.LEADING)
                            .addComponent(paddleSwapCheckBox)
                            .addComponent(autospaceCheckBox))
                        .addGap(18, 18, 18)
                        .addGroup(cwControlPanelLayout.createParallelGroup(javax.swing.GroupLayout.Alignment.LEADING)
                            .addComponent(contestSpacingCheckBox)
                            .addComponent(paddleWatchDogDisableCheckBox)))
                    .addGroup(cwControlPanelLayout.createSequentialGroup()
                        .addComponent(jLabel7)
                        .addPreferredGap(javax.swing.LayoutStyle.ComponentPlacement.RELATED)
                        .addComponent(sideToneFreqComboBox, javax.swing.GroupLayout.PREFERRED_SIZE, javax.swing.GroupLayout.DEFAULT_SIZE, javax.swing.GroupLayout.PREFERRED_SIZE))
                    .addGroup(cwControlPanelLayout.createSequentialGroup()
                        .addComponent(sidetoneEnableCheckBox)
                        .addGap(18, 18, 18)
                        .addComponent(paddleOnlySidetoneCheckBox)))
                .addGap(0, 20, Short.MAX_VALUE))
        );
        cwControlPanelLayout.setVerticalGroup(
            cwControlPanelLayout.createParallelGroup(javax.swing.GroupLayout.Alignment.LEADING)
            .addGroup(cwControlPanelLayout.createSequentialGroup()
                .addContainerGap()
                .addGroup(cwControlPanelLayout.createParallelGroup(javax.swing.GroupLayout.Alignment.BASELINE)
                    .addComponent(jLabel17)
                    .addComponent(dahDitRatioSpinner, javax.swing.GroupLayout.PREFERRED_SIZE, javax.swing.GroupLayout.DEFAULT_SIZE, javax.swing.GroupLayout.PREFERRED_SIZE))
                .addPreferredGap(javax.swing.LayoutStyle.ComponentPlacement.RELATED)
                .addGroup(cwControlPanelLayout.createParallelGroup(javax.swing.GroupLayout.Alignment.BASELINE)
                    .addComponent(jLabel8)
                    .addComponent(weightingSpinner, javax.swing.GroupLayout.PREFERRED_SIZE, javax.swing.GroupLayout.DEFAULT_SIZE, javax.swing.GroupLayout.PREFERRED_SIZE))
                .addPreferredGap(javax.swing.LayoutStyle.ComponentPlacement.RELATED)
                .addGroup(cwControlPanelLayout.createParallelGroup(javax.swing.GroupLayout.Alignment.BASELINE)
                    .addComponent(farnsWpmSpinner, javax.swing.GroupLayout.PREFERRED_SIZE, javax.swing.GroupLayout.DEFAULT_SIZE, javax.swing.GroupLayout.PREFERRED_SIZE)
                    .addComponent(jLabel12))
                .addPreferredGap(javax.swing.LayoutStyle.ComponentPlacement.RELATED)
                .addGroup(cwControlPanelLayout.createParallelGroup(javax.swing.GroupLayout.Alignment.BASELINE)
                    .addComponent(keyModeComboBox, javax.swing.GroupLayout.PREFERRED_SIZE, javax.swing.GroupLayout.DEFAULT_SIZE, javax.swing.GroupLayout.PREFERRED_SIZE)
                    .addComponent(jLabel13))
                .addPreferredGap(javax.swing.LayoutStyle.ComponentPlacement.RELATED)
                .addGroup(cwControlPanelLayout.createParallelGroup(javax.swing.GroupLayout.Alignment.BASELINE)
                    .addComponent(paddleSwitchpointSpinner, javax.swing.GroupLayout.PREFERRED_SIZE, javax.swing.GroupLayout.DEFAULT_SIZE, javax.swing.GroupLayout.PREFERRED_SIZE)
                    .addComponent(jLabel16))
                .addPreferredGap(javax.swing.LayoutStyle.ComponentPlacement.UNRELATED)
                .addGroup(cwControlPanelLayout.createParallelGroup(javax.swing.GroupLayout.Alignment.BASELINE)
                    .addComponent(autospaceCheckBox)
                    .addComponent(contestSpacingCheckBox))
                .addPreferredGap(javax.swing.LayoutStyle.ComponentPlacement.RELATED)
                .addGroup(cwControlPanelLayout.createParallelGroup(javax.swing.GroupLayout.Alignment.BASELINE)
                    .addComponent(paddleSwapCheckBox)
                    .addComponent(paddleWatchDogDisableCheckBox))
                .addPreferredGap(javax.swing.LayoutStyle.ComponentPlacement.UNRELATED)
                .addGroup(cwControlPanelLayout.createParallelGroup(javax.swing.GroupLayout.Alignment.BASELINE)
                    .addComponent(jLabel7)
                    .addComponent(sideToneFreqComboBox, javax.swing.GroupLayout.PREFERRED_SIZE, javax.swing.GroupLayout.DEFAULT_SIZE, javax.swing.GroupLayout.PREFERRED_SIZE))
                .addPreferredGap(javax.swing.LayoutStyle.ComponentPlacement.UNRELATED)
                .addGroup(cwControlPanelLayout.createParallelGroup(javax.swing.GroupLayout.Alignment.BASELINE)
                    .addComponent(sidetoneEnableCheckBox)
                    .addComponent(paddleOnlySidetoneCheckBox))
                .addContainerGap(110, Short.MAX_VALUE))
        );

        jTabbedPane1.addTab("CW Control", cwControlPanel);

        pttTailSpinner.setModel(new javax.swing.SpinnerNumberModel(0, 0, 250, 10));
        pttTailSpinner.addChangeListener(new javax.swing.event.ChangeListener() {
            public void stateChanged(javax.swing.event.ChangeEvent evt) {
                pttTailSpinnerStateChanged(evt);
            }
        });

        pttLeadSpinner.setModel(new javax.swing.SpinnerNumberModel(0, 0, 250, 10));
        pttLeadSpinner.addChangeListener(new javax.swing.event.ChangeListener() {
            public void stateChanged(javax.swing.event.ChangeEvent evt) {
                pttLeadSpinnerStateChanged(evt);
            }
        });

        jLabel9.setText("PTT Lead [ms]");

        jLabel10.setText("PTT Tail [ms]");

        pttEnableCheckBox.setText("PTT Enable");
        pttEnableCheckBox.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                pttEnableCheckBoxActionPerformed(evt);
            }
        });

        keyOut1CheckBox.setText("Key Out 1 Enable");
        keyOut1CheckBox.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                keyOut1CheckBoxActionPerformed(evt);
            }
        });

        keyOut2CheckBox.setText("Key Out 2 Enable");
        keyOut2CheckBox.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                keyOut2CheckBoxActionPerformed(evt);
            }
        });

        hangTimeComboBox.setModel(new javax.swing.DefaultComboBoxModel(new String[] { "1 wordspace + 1 dit", "1 wordspace + 2 dits", "1 wordspace + 4 dits", "1 wordspace + 8 dits" }));
        hangTimeComboBox.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                hangTimeComboBoxActionPerformed(evt);
            }
        });

        jLabel11.setText("Hang Time");

        jLabel15.setText("Key Compensation [ms]");

        keyCompSpinner.setModel(new javax.swing.SpinnerNumberModel(0, 0, 250, 1));
        keyCompSpinner.addChangeListener(new javax.swing.event.ChangeListener() {
            public void stateChanged(javax.swing.event.ChangeEvent evt) {
                keyCompSpinnerStateChanged(evt);
            }
        });

        jLabel14.setText("1st Extension [ms]");

        firstExtnSpinner.setModel(new javax.swing.SpinnerNumberModel(0, 0, 250, 1));
        firstExtnSpinner.addChangeListener(new javax.swing.event.ChangeListener() {
            public void stateChanged(javax.swing.event.ChangeEvent evt) {
                firstExtnSpinnerStateChanged(evt);
            }
        });

        jPanel2.setBorder(javax.swing.BorderFactory.createTitledBorder(javax.swing.BorderFactory.createTitledBorder(""), "cwdaemon PTT"));

        pttButtonGroup.add(controlsPttRadioButton);
        controlsPttRadioButton.setText("Controls PTT");
        controlsPttRadioButton.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                controlsPttRadioButtonActionPerformed(evt);
            }
        });

        pttButtonGroup.add(controlsKeyRadioButton);
        controlsKeyRadioButton.setText("Controls Key Out");
        controlsKeyRadioButton.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                controlsKeyRadioButtonActionPerformed(evt);
            }
        });

        javax.swing.GroupLayout jPanel2Layout = new javax.swing.GroupLayout(jPanel2);
        jPanel2.setLayout(jPanel2Layout);
        jPanel2Layout.setHorizontalGroup(
            jPanel2Layout.createParallelGroup(javax.swing.GroupLayout.Alignment.LEADING)
            .addComponent(controlsPttRadioButton)
            .addComponent(controlsKeyRadioButton)
        );
        jPanel2Layout.setVerticalGroup(
            jPanel2Layout.createParallelGroup(javax.swing.GroupLayout.Alignment.LEADING)
            .addGroup(jPanel2Layout.createSequentialGroup()
                .addComponent(controlsPttRadioButton)
                .addPreferredGap(javax.swing.LayoutStyle.ComponentPlacement.RELATED)
                .addComponent(controlsKeyRadioButton))
        );

        javax.swing.GroupLayout txControlPanelLayout = new javax.swing.GroupLayout(txControlPanel);
        txControlPanel.setLayout(txControlPanelLayout);
        txControlPanelLayout.setHorizontalGroup(
            txControlPanelLayout.createParallelGroup(javax.swing.GroupLayout.Alignment.LEADING)
            .addGroup(txControlPanelLayout.createSequentialGroup()
                .addContainerGap()
                .addGroup(txControlPanelLayout.createParallelGroup(javax.swing.GroupLayout.Alignment.LEADING)
                    .addGroup(txControlPanelLayout.createParallelGroup(javax.swing.GroupLayout.Alignment.TRAILING, false)
                        .addGroup(txControlPanelLayout.createSequentialGroup()
                            .addGroup(txControlPanelLayout.createParallelGroup(javax.swing.GroupLayout.Alignment.TRAILING)
                                .addGroup(txControlPanelLayout.createSequentialGroup()
                                    .addGroup(txControlPanelLayout.createParallelGroup(javax.swing.GroupLayout.Alignment.LEADING)
                                        .addComponent(jLabel10)
                                        .addComponent(jLabel11))
                                    .addGap(26, 26, 26))
                                .addGroup(javax.swing.GroupLayout.Alignment.LEADING, txControlPanelLayout.createSequentialGroup()
                                    .addComponent(jLabel9)
                                    .addPreferredGap(javax.swing.LayoutStyle.ComponentPlacement.RELATED)))
                            .addGroup(txControlPanelLayout.createParallelGroup(javax.swing.GroupLayout.Alignment.LEADING)
                                .addComponent(pttLeadSpinner, javax.swing.GroupLayout.PREFERRED_SIZE, javax.swing.GroupLayout.DEFAULT_SIZE, javax.swing.GroupLayout.PREFERRED_SIZE)
                                .addComponent(pttTailSpinner, javax.swing.GroupLayout.PREFERRED_SIZE, javax.swing.GroupLayout.DEFAULT_SIZE, javax.swing.GroupLayout.PREFERRED_SIZE)
                                .addComponent(hangTimeComboBox, javax.swing.GroupLayout.PREFERRED_SIZE, javax.swing.GroupLayout.DEFAULT_SIZE, javax.swing.GroupLayout.PREFERRED_SIZE))
                            .addGap(25, 25, 25))
                        .addGroup(javax.swing.GroupLayout.Alignment.LEADING, txControlPanelLayout.createSequentialGroup()
                            .addGroup(txControlPanelLayout.createParallelGroup(javax.swing.GroupLayout.Alignment.LEADING)
                                .addComponent(jLabel15)
                                .addComponent(jLabel14))
                            .addPreferredGap(javax.swing.LayoutStyle.ComponentPlacement.RELATED)
                            .addGroup(txControlPanelLayout.createParallelGroup(javax.swing.GroupLayout.Alignment.LEADING)
                                .addComponent(firstExtnSpinner, javax.swing.GroupLayout.PREFERRED_SIZE, javax.swing.GroupLayout.DEFAULT_SIZE, javax.swing.GroupLayout.PREFERRED_SIZE)
                                .addComponent(keyCompSpinner, javax.swing.GroupLayout.PREFERRED_SIZE, javax.swing.GroupLayout.DEFAULT_SIZE, javax.swing.GroupLayout.PREFERRED_SIZE))))
                    .addGroup(txControlPanelLayout.createSequentialGroup()
                        .addComponent(jPanel2, javax.swing.GroupLayout.PREFERRED_SIZE, javax.swing.GroupLayout.DEFAULT_SIZE, javax.swing.GroupLayout.PREFERRED_SIZE)
                        .addGroup(txControlPanelLayout.createParallelGroup(javax.swing.GroupLayout.Alignment.LEADING)
                            .addGroup(txControlPanelLayout.createSequentialGroup()
                                .addGap(19, 19, 19)
                                .addGroup(txControlPanelLayout.createParallelGroup(javax.swing.GroupLayout.Alignment.LEADING)
                                    .addComponent(keyOut1CheckBox)
                                    .addComponent(pttEnableCheckBox)))
                            .addGroup(txControlPanelLayout.createSequentialGroup()
                                .addGap(18, 18, 18)
                                .addComponent(keyOut2CheckBox)))))
                .addContainerGap(19, Short.MAX_VALUE))
        );
        txControlPanelLayout.setVerticalGroup(
            txControlPanelLayout.createParallelGroup(javax.swing.GroupLayout.Alignment.LEADING)
            .addGroup(txControlPanelLayout.createSequentialGroup()
                .addContainerGap()
                .addGroup(txControlPanelLayout.createParallelGroup(javax.swing.GroupLayout.Alignment.LEADING)
                    .addGroup(txControlPanelLayout.createSequentialGroup()
                        .addGap(12, 12, 12)
                        .addComponent(pttEnableCheckBox)
                        .addPreferredGap(javax.swing.LayoutStyle.ComponentPlacement.RELATED)
                        .addComponent(keyOut1CheckBox)
                        .addPreferredGap(javax.swing.LayoutStyle.ComponentPlacement.RELATED)
                        .addComponent(keyOut2CheckBox))
                    .addComponent(jPanel2, javax.swing.GroupLayout.PREFERRED_SIZE, javax.swing.GroupLayout.DEFAULT_SIZE, javax.swing.GroupLayout.PREFERRED_SIZE))
                .addPreferredGap(javax.swing.LayoutStyle.ComponentPlacement.RELATED)
                .addGroup(txControlPanelLayout.createParallelGroup(javax.swing.GroupLayout.Alignment.LEADING)
                    .addComponent(jLabel9)
                    .addComponent(pttLeadSpinner, javax.swing.GroupLayout.PREFERRED_SIZE, javax.swing.GroupLayout.DEFAULT_SIZE, javax.swing.GroupLayout.PREFERRED_SIZE))
                .addPreferredGap(javax.swing.LayoutStyle.ComponentPlacement.RELATED)
                .addGroup(txControlPanelLayout.createParallelGroup(javax.swing.GroupLayout.Alignment.BASELINE)
                    .addComponent(pttTailSpinner, javax.swing.GroupLayout.PREFERRED_SIZE, javax.swing.GroupLayout.DEFAULT_SIZE, javax.swing.GroupLayout.PREFERRED_SIZE)
                    .addComponent(jLabel10))
                .addPreferredGap(javax.swing.LayoutStyle.ComponentPlacement.RELATED)
                .addGroup(txControlPanelLayout.createParallelGroup(javax.swing.GroupLayout.Alignment.BASELINE)
                    .addComponent(hangTimeComboBox, javax.swing.GroupLayout.PREFERRED_SIZE, javax.swing.GroupLayout.DEFAULT_SIZE, javax.swing.GroupLayout.PREFERRED_SIZE)
                    .addComponent(jLabel11))
                .addPreferredGap(javax.swing.LayoutStyle.ComponentPlacement.RELATED)
                .addGroup(txControlPanelLayout.createParallelGroup(javax.swing.GroupLayout.Alignment.BASELINE)
                    .addComponent(keyCompSpinner, javax.swing.GroupLayout.PREFERRED_SIZE, javax.swing.GroupLayout.DEFAULT_SIZE, javax.swing.GroupLayout.PREFERRED_SIZE)
                    .addComponent(jLabel15))
                .addPreferredGap(javax.swing.LayoutStyle.ComponentPlacement.RELATED)
                .addGroup(txControlPanelLayout.createParallelGroup(javax.swing.GroupLayout.Alignment.BASELINE)
                    .addComponent(firstExtnSpinner, javax.swing.GroupLayout.PREFERRED_SIZE, javax.swing.GroupLayout.DEFAULT_SIZE, javax.swing.GroupLayout.PREFERRED_SIZE)
                    .addComponent(jLabel14))
                .addContainerGap(153, Short.MAX_VALUE))
        );

        jTabbedPane1.addTab("Tx Control", txControlPanel);

        javax.swing.GroupLayout layout = new javax.swing.GroupLayout(getContentPane());
        getContentPane().setLayout(layout);
        layout.setHorizontalGroup(
            layout.createParallelGroup(javax.swing.GroupLayout.Alignment.LEADING)
            .addComponent(jTabbedPane1, javax.swing.GroupLayout.PREFERRED_SIZE, 361, javax.swing.GroupLayout.PREFERRED_SIZE)
        );
        layout.setVerticalGroup(
            layout.createParallelGroup(javax.swing.GroupLayout.Alignment.LEADING)
            .addComponent(jTabbedPane1, javax.swing.GroupLayout.PREFERRED_SIZE, 500, javax.swing.GroupLayout.PREFERRED_SIZE)
        );

        pack();
    }// </editor-fold>//GEN-END:initComponents

    private void hostSetsSpeedButtonActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_hostSetsSpeedButtonActionPerformed
        if (!initialising) {
            s.HOST_SETS_SPEED = "true";
            hostWpmTextField.setEnabled(true);
            minWpmTextField.setEnabled(false);
            maxWpmTextField.setEnabled(false);
            potWpmTextField.setEnabled(false);
            WINKEY_WPM = s.HOST_WPM;
            wsc.setWpmSpeed((byte) Integer.parseInt(WINKEY_WPM, 16));
        }
    }//GEN-LAST:event_hostSetsSpeedButtonActionPerformed

    private void cwdaemonPortTextFieldKeyReleased(java.awt.event.KeyEvent evt) {//GEN-FIRST:event_cwdaemonPortTextFieldKeyReleased
        if (!initialising) {
            s.CWDAEMON_PORT = cwdaemonPortTextField.getText();
            udpServer.stopServer();
            udpServer.startServer(Integer.parseInt(s.CWDAEMON_PORT));
            new Thread(udpServer).start();
        }
    }//GEN-LAST:event_cwdaemonPortTextFieldKeyReleased

    private void hostWpmTextFieldKeyReleased(java.awt.event.KeyEvent evt) {//GEN-FIRST:event_hostWpmTextFieldKeyReleased
        // Delegate to method without KeyEvent passed to it for convenience.
        hostWpmTextFieldChanged();
    }//GEN-LAST:event_hostWpmTextFieldKeyReleased

    // Delegate method for hostWpmTextFieldKeyReleased to simplify other access.
    private void hostWpmTextFieldChanged() {
        if (!initialising) {
            if (hostWpmTextField.getText().matches("\\d+")) {
                s.HOST_WPM = Integer.toHexString(Integer.parseInt(hostWpmTextField.getText()));
                // If the host is in control, send the new speed to winkey.
                if (s.HOST_SETS_SPEED.equalsIgnoreCase("true")) {
                    WINKEY_WPM = s.HOST_WPM;
                    wsc.setWpmSpeed((byte) Integer.parseInt(WINKEY_WPM, 16));
                }
            }
        }
    } // hostWpmTextFieldChanged.
    
    
    private void serialPortComboBoxActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_serialPortComboBoxActionPerformed
        if (!initialising) {
            s.SERIAL_PORT = serialPortComboBox.getSelectedItem().toString();
            // Note this does NOT capture manual edits without hitting <enter>.
            // Adding this functionality would require a document listener.
            // For now let's ignore this requirement.
        }
    }//GEN-LAST:event_serialPortComboBoxActionPerformed

    private void potSetsSpeedButtonActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_potSetsSpeedButtonActionPerformed
        if (!initialising) {
            s.HOST_SETS_SPEED = "false";
            hostWpmTextField.setEnabled(false);
            minWpmTextField.setEnabled(true);
            maxWpmTextField.setEnabled(true);
            potWpmTextField.setEnabled(true);
            WINKEY_WPM = "00";
            wsc.setWpmSpeed((byte) Integer.parseInt(WINKEY_WPM, 16));
        }
    }//GEN-LAST:event_potSetsSpeedButtonActionPerformed

    private void minWpmTextFieldKeyReleased(java.awt.event.KeyEvent evt) {//GEN-FIRST:event_minWpmTextFieldKeyReleased
        if (minWpmTextField.getText().matches("\\d+")) {
            // Trap disallowed case where it's negative.
            if (Integer.parseInt(maxWpmTextField.getText())
                    - Integer.parseInt(minWpmTextField.getText()) > 0) {
                // Winkey needs the range, not the max:
                s.WINKEY_MIN_WPM = Integer.toHexString(
                        Integer.parseInt(minWpmTextField.getText()));
                wsc.setUpSpeedPot((byte) Integer.parseInt(s.WINKEY_MIN_WPM, 16),
                        (byte) Integer.parseInt(s.WINKEY_WPM_RANGE, 16));
                // Force an immediate speed pot update.
                wsc.setWpmSpeed((byte) 0x00);
                // Issue read of speed pot request to update display.
                wsc.getWpmSpeed();
            }
        }
    }//GEN-LAST:event_minWpmTextFieldKeyReleased

    private void maxWpmTextFieldKeyReleased(java.awt.event.KeyEvent evt) {//GEN-FIRST:event_maxWpmTextFieldKeyReleased
        if (maxWpmTextField.getText().matches("\\d+")) {
            // Trap disallowed case where it's negative.
            if (Integer.parseInt(maxWpmTextField.getText())
                    - Integer.parseInt(minWpmTextField.getText()) > 0) {
                // Winkey needs the range, not the max:
                s.WINKEY_WPM_RANGE = Integer.toHexString(
                        Integer.parseInt(maxWpmTextField.getText())
                        - Integer.parseInt(s.WINKEY_MIN_WPM, 16));
                wsc.setUpSpeedPot((byte) Integer.parseInt(s.WINKEY_MIN_WPM, 16),
                        (byte) Integer.parseInt(s.WINKEY_WPM_RANGE, 16));
                // Force an immediate speed pot update.
                wsc.setWpmSpeed((byte) 0x00);
                // Issue read of speed pot request to update display.
                wsc.getWpmSpeed();
            }
        }
    }//GEN-LAST:event_maxWpmTextFieldKeyReleased

    private void dahDitRatioSpinnerStateChanged(javax.swing.event.ChangeEvent evt) {//GEN-FIRST:event_dahDitRatioSpinnerStateChanged
        // Convert the displayed value to the code required for Winkey.
        if (!initialising) {
            float dahDitRatio = Float.parseFloat(dahDitRatioSpinner.getValue().toString());
            int dahDitCode = Math.round(50*dahDitRatio/3);
            // Overwrite saved settings with new value.
            s.WINKEY_DAH_DIT_RATIO = Integer.toHexString(dahDitCode);
            // Send new setting to Winkey.
            wsc.setDahDitRatio((byte) Integer.parseInt(s.WINKEY_DAH_DIT_RATIO, 16));
        }
    }//GEN-LAST:event_dahDitRatioSpinnerStateChanged

    private void weightingSpinnerStateChanged(javax.swing.event.ChangeEvent evt) {//GEN-FIRST:event_weightingSpinnerStateChanged
        // Delegate to method without KeyEvent passed to it for convenience.
        weightingSpinnerChanged();
    }//GEN-LAST:event_weightingSpinnerStateChanged

    // Delegate method for weightingSpinnerStateChanged to simplify other access.
    private void weightingSpinnerChanged() {                                              
        // Convert the displayed value to the code required for Winkey.
        if (!initialising) {
            int weight = Integer.parseInt(weightingSpinner.getValue().toString());
            // Overwrite saved settings with new value.
            s.WINKEY_WEIGHT = Integer.toHexString(weight);
            // Send new setting to Winkey.
            wsc.setWeighting((byte) Integer.parseInt(s.WINKEY_WEIGHT, 16));
        }
    } // weightingSpinnerChanged
    
    
    private void farnsWpmSpinnerStateChanged(javax.swing.event.ChangeEvent evt) {//GEN-FIRST:event_farnsWpmSpinnerStateChanged
        // Convert the displayed value to the code required for Winkey.
        if (!initialising) {
            int farnsWpm = Integer.parseInt(farnsWpmSpinner.getValue().toString());
            // Overwrite saved settings with new value.
            s.WINKEY_WPM_FARNS = Integer.toHexString(farnsWpm);
            // Send new setting to Winkey.
            wsc.setFarnsWpm((byte) Integer.parseInt(s.WINKEY_WPM_FARNS, 16));
        }
    }//GEN-LAST:event_farnsWpmSpinnerStateChanged

    private void keyModeComboBoxActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_keyModeComboBoxActionPerformed
        // Two actions are required: decode the key mode, and the paddle priority.
        if (!initialising) {
            int index = keyModeComboBox.getSelectedIndex();
            int winkeyMode;
            int winkeyPinConfig;
            int keyMode;
            int priority;
        
            // Get the previous settings but AND-out the key mode and priority parts.
            int previousWinkeyMode = 0xCF & Integer.parseInt(s.WINKEY_MODE, 16);
            int previousWinkeyPinConfig = 0x3F & Integer.parseInt(s.WINKEY_PIN_CONFIG, 16);
            
            if (index < 2) {
                // Iambic modes.
                keyMode = index;
                // Shift to upper nibble and combine with other previous settings.
                winkeyMode = previousWinkeyMode | (keyMode << 4);
                // Overwrite saved settings with new value.
                s.WINKEY_MODE = Integer.toHexString(winkeyMode);
                // Send new setting to Winkey.
                wsc.setWinkeyerMode((byte) Integer.parseInt(s.WINKEY_MODE, 16));
            } else if (index < 5) {
                // Ultimatic modes.
                // Update WINKEY_MODE first...
                keyMode = 0x2;
                // Shift to upper nibble and combine with other previous settings.
                winkeyMode = previousWinkeyMode | (keyMode << 4);
                // Overwrite saved settings with new value.
                s.WINKEY_MODE = Integer.toHexString(winkeyMode);
                // Send new setting to Winkey.
                wsc.setWinkeyerMode((byte) Integer.parseInt(s.WINKEY_MODE, 16));
                // Now update WINKEY_PIN_CONFIG...
                priority = index - 2;
                // Shift to bits 6/7 and combine with other previous settings.
                winkeyPinConfig = previousWinkeyPinConfig | (priority << 6);
                // Overwrite saved settings with new value.
                s.WINKEY_PIN_CONFIG = Integer.toHexString(winkeyPinConfig);
                // Send new setting to Winkey.
                wsc.setPinConfig((byte) Integer.parseInt(s.WINKEY_PIN_CONFIG, 16));
            } else {
                // Bug mode.
                keyMode = 0x3;
                // Shift to upper nibble and combine with other previous settings.
                winkeyMode = previousWinkeyMode | (keyMode << 4);
                // Overwrite saved settings with new value.
                s.WINKEY_MODE = Integer.toHexString(winkeyMode);
                // Send new setting to Winkey.
                wsc.setWinkeyerMode((byte) Integer.parseInt(s.WINKEY_MODE, 16));
            }       
        }
    }//GEN-LAST:event_keyModeComboBoxActionPerformed

    private void paddleSwitchpointSpinnerStateChanged(javax.swing.event.ChangeEvent evt) {//GEN-FIRST:event_paddleSwitchpointSpinnerStateChanged
        // Convert the displayed value to the code required for Winkey.
        if (!initialising) {
            int paddleSwitchpoint = Integer.parseInt(paddleSwitchpointSpinner.getValue().toString());
            // Overwrite saved settings with new value.
            s.WINKEY_PADDLE_SWITCHPOINT = Integer.toHexString(paddleSwitchpoint);
            // Send new setting to Winkey.
            wsc.setPaddleSwitchpoint((byte) Integer.parseInt(s.WINKEY_PADDLE_SWITCHPOINT, 16));
        }
    }//GEN-LAST:event_paddleSwitchpointSpinnerStateChanged

    private void autospaceCheckBoxActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_autospaceCheckBoxActionPerformed
        // Convert the displayed setting to the code required for Winkey.
        if (!initialising) {
            
            int winkeyMode;
            // Get the previous settings but AND-out the autospace part.
            int previousWinkeyMode = 0xFD & Integer.parseInt(s.WINKEY_MODE, 16);
            
            if (autospaceCheckBox.isSelected()) {
                winkeyMode = previousWinkeyMode | 0x02;
            } else {
                winkeyMode = previousWinkeyMode | 0x00;
            }
            // Overwrite saved settings with new value.
            s.WINKEY_MODE = Integer.toHexString(winkeyMode);
            // Send new setting to Winkey.
            wsc.setWinkeyerMode((byte) Integer.parseInt(s.WINKEY_MODE, 16));
            
        }
    }//GEN-LAST:event_autospaceCheckBoxActionPerformed

    private void contestSpacingCheckBoxActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_contestSpacingCheckBoxActionPerformed
        // Convert the displayed setting to the code required for Winkey.
        if (!initialising) {
            
            int winkeyMode;
            // Get the previous settings but AND-out the contest spacing part.
            int previousWinkeyMode = 0xFE & Integer.parseInt(s.WINKEY_MODE, 16);
            
            if (contestSpacingCheckBox.isSelected()) {
                winkeyMode = previousWinkeyMode | 0x01;
            } else {
                winkeyMode = previousWinkeyMode | 0x00;
            }
            // Overwrite saved settings with new value.
            s.WINKEY_MODE = Integer.toHexString(winkeyMode);
            // Send new setting to Winkey.
            wsc.setWinkeyerMode((byte) Integer.parseInt(s.WINKEY_MODE, 16));
            
        }
    }//GEN-LAST:event_contestSpacingCheckBoxActionPerformed

    private void paddleSwapCheckBoxActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_paddleSwapCheckBoxActionPerformed
        // Convert the displayed setting to the code required for Winkey.
        if (!initialising) {
            
            int winkeyMode;
            // Get the previous settings but AND-out the paddle swap part.
            int previousWinkeyMode = 0xF7 & Integer.parseInt(s.WINKEY_MODE, 16);
            
            if (paddleSwapCheckBox.isSelected()) {
                winkeyMode = previousWinkeyMode | 0x08;
            } else {
                winkeyMode = previousWinkeyMode | 0x00;
            }
            // Overwrite saved settings with new value.
            s.WINKEY_MODE = Integer.toHexString(winkeyMode);
            // Send new setting to Winkey.
            wsc.setWinkeyerMode((byte) Integer.parseInt(s.WINKEY_MODE, 16));
            
        }
    }//GEN-LAST:event_paddleSwapCheckBoxActionPerformed

    private void paddleWatchDogDisableCheckBoxActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_paddleWatchDogDisableCheckBoxActionPerformed
        // Convert the displayed setting to the code required for Winkey.
        if (!initialising) {
            
            int winkeyMode;
            // Get the previous settings but AND-out the paddle watchdog part.
            int previousWinkeyMode = 0x7F & Integer.parseInt(s.WINKEY_MODE, 16);
            
            if (paddleWatchDogDisableCheckBox.isSelected()) {
                winkeyMode = previousWinkeyMode | 0x80;
            } else {
                winkeyMode = previousWinkeyMode | 0x00;
            }
            // Overwrite saved settings with new value.
            s.WINKEY_MODE = Integer.toHexString(winkeyMode);
            // Send new setting to Winkey.
            wsc.setWinkeyerMode((byte) Integer.parseInt(s.WINKEY_MODE, 16));
            
        }
    }//GEN-LAST:event_paddleWatchDogDisableCheckBoxActionPerformed

    private void sideToneFreqComboBoxActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_sideToneFreqComboBoxActionPerformed
        // Delegate to method without KeyEvent passed to it for convenience.
        sideToneFreqComboBoxChanged();
    }//GEN-LAST:event_sideToneFreqComboBoxActionPerformed

    // Delegate method for sidetoneEnableCheckBoxActionPerformed to simplify other access.
    private void sideToneFreqComboBoxChanged() {
        // Convert the displayed setting to the code required for Winkey.
        if (!initialising) {
            
            int index = sideToneFreqComboBox.getSelectedIndex();
            
            // Get the previous settings but AND-out the sidetone frequency part.
            int previousSidetone = 0xF0 & Integer.parseInt(s.WINKEY_SIDETONE, 16);
            
            int sidetone = previousSidetone | (10 - index);
            
            // Overwrite saved settings with new value.
            s.WINKEY_SIDETONE = Integer.toHexString(sidetone);
            // Send new setting to Winkey.
            wsc.setSidetone((byte) Integer.parseInt(s.WINKEY_SIDETONE, 16));
            
        }
    } // sideToneFreqComboBoxChanged.
    
    
    private void paddleOnlySidetoneCheckBoxActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_paddleOnlySidetoneCheckBoxActionPerformed
        // Convert the displayed setting to the code required for Winkey.
        if (!initialising) {
            
            int sidetone;
            // Get the previous settings but AND-out the paddle only sidetone part.
            int previousSidetone = 0x7F & Integer.parseInt(s.WINKEY_SIDETONE, 16);
            
            if (paddleOnlySidetoneCheckBox.isSelected()) {
                sidetone = previousSidetone | 0x80;
            } else {
                sidetone = previousSidetone | 0x00;
            }
            // Overwrite saved settings with new value.
            s.WINKEY_SIDETONE = Integer.toHexString(sidetone);
            // Send new setting to Winkey.
            wsc.setSidetone((byte) Integer.parseInt(s.WINKEY_SIDETONE, 16));
            
        }
    }//GEN-LAST:event_paddleOnlySidetoneCheckBoxActionPerformed

    private void sidetoneEnableCheckBoxActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_sidetoneEnableCheckBoxActionPerformed
        // Delegate to method without KeyEvent passed to it for convenience.
        sidetoneEnableCheckBoxChanged();
    }//GEN-LAST:event_sidetoneEnableCheckBoxActionPerformed

    // Delegate method for sidetoneEnableCheckBoxActionPerformed to simplify other access.
    private void sidetoneEnableCheckBoxChanged() {
        // Convert the displayed setting to the code required for Winkey.
        if (!initialising) {
            
            int pinConfig;
            // Get the previous settings but AND-out the sidetone enable part.
            int previousPinConfig = 0xFD & Integer.parseInt(s.WINKEY_PIN_CONFIG, 16);
            
            if (sidetoneEnableCheckBox.isSelected()) {
                pinConfig = previousPinConfig | 0x02;
            } else {
                pinConfig = previousPinConfig | 0x00;
            }
            // Overwrite saved settings with new value.
            s.WINKEY_PIN_CONFIG = Integer.toHexString(pinConfig);
            // Send new setting to Winkey.
            wsc.setPinConfig((byte) Integer.parseInt(s.WINKEY_PIN_CONFIG, 16));
            
        }
    } // sidetoneEnableCheckBoxChanged
    
    
    private void pttEnableCheckBoxActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_pttEnableCheckBoxActionPerformed
        // Convert the displayed setting to the code required for Winkey.
        if (!initialising) {
            
            int pinConfig;
            // Get the previous settings but AND-out the ptt enable part.
            int previousPinConfig = 0xFE & Integer.parseInt(s.WINKEY_PIN_CONFIG, 16);
            
            if (pttEnableCheckBox.isSelected()) {
                pinConfig = previousPinConfig | 0x01;
            } else {
                pinConfig = previousPinConfig | 0x00;
            }
            // Overwrite saved settings with new value.
            s.WINKEY_PIN_CONFIG = Integer.toHexString(pinConfig);
            // Send new setting to Winkey.
            wsc.setPinConfig((byte) Integer.parseInt(s.WINKEY_PIN_CONFIG, 16));
            
        }
    }//GEN-LAST:event_pttEnableCheckBoxActionPerformed

    private void keyOut1CheckBoxActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_keyOut1CheckBoxActionPerformed
        // Convert the displayed setting to the code required for Winkey.
        if (!initialising) {
            
            int pinConfig;
            // Get the previous settings but AND-out the keyOut1 enable part.
            int previousPinConfig = 0xF7 & Integer.parseInt(s.WINKEY_PIN_CONFIG, 16);
            
            if (keyOut1CheckBox.isSelected()) {
                pinConfig = previousPinConfig | 0x08;
            } else {
                pinConfig = previousPinConfig | 0x00;
            }
            // Overwrite saved settings with new value.
            s.WINKEY_PIN_CONFIG = Integer.toHexString(pinConfig);
            // Send new setting to Winkey.
            wsc.setPinConfig((byte) Integer.parseInt(s.WINKEY_PIN_CONFIG, 16));
            
        }
    }//GEN-LAST:event_keyOut1CheckBoxActionPerformed

    private void keyOut2CheckBoxActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_keyOut2CheckBoxActionPerformed
        // Convert the displayed setting to the code required for Winkey.
        if (!initialising) {
            
            int pinConfig;
            // Get the previous settings but AND-out the keyOut2 enable part.
            int previousPinConfig = 0xFB & Integer.parseInt(s.WINKEY_PIN_CONFIG, 16);
            
            if (keyOut2CheckBox.isSelected()) {
                pinConfig = previousPinConfig | 0x04;
            } else {
                pinConfig = previousPinConfig | 0x00;
            }
            // Overwrite saved settings with new value.
            s.WINKEY_PIN_CONFIG = Integer.toHexString(pinConfig);
            // Send new setting to Winkey.
            wsc.setPinConfig((byte) Integer.parseInt(s.WINKEY_PIN_CONFIG, 16));
            
        }
    }//GEN-LAST:event_keyOut2CheckBoxActionPerformed

    private void pttTailSpinnerStateChanged(javax.swing.event.ChangeEvent evt) {//GEN-FIRST:event_pttTailSpinnerStateChanged
        // Convert the displayed value to the code required for Winkey.
        if (!initialising) {
            
            int pttLead = Integer.parseInt(pttTailSpinner.getValue().toString());
            // Overwrite saved settings with new value.
            s.WINKEY_PTT_TAIL = Integer.toHexString(pttLead);
            // Send new setting to Winkey.
            wsc.setPttLeadTail((byte) Integer.parseInt(s.WINKEY_PTT_LEAD, 16),
                            (byte) Integer.parseInt(s.WINKEY_PTT_TAIL, 16));
        }
    }//GEN-LAST:event_pttTailSpinnerStateChanged

    private void pttLeadSpinnerStateChanged(javax.swing.event.ChangeEvent evt) {//GEN-FIRST:event_pttLeadSpinnerStateChanged
        // Convert the displayed value to the code required for Winkey.
        if (!initialising) {
            
            int pttLead = Integer.parseInt(pttLeadSpinner.getValue().toString());
            // Overwrite saved settings with new value.
            s.WINKEY_PTT_LEAD = Integer.toHexString(pttLead);
            // Send new setting to Winkey.
            wsc.setPttLeadTail((byte) Integer.parseInt(s.WINKEY_PTT_LEAD, 16),
                            (byte) Integer.parseInt(s.WINKEY_PTT_TAIL, 16));
        }
    }//GEN-LAST:event_pttLeadSpinnerStateChanged

    private void hangTimeComboBoxActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_hangTimeComboBoxActionPerformed
        // Convert the displayed setting to the code required for Winkey.
        if (!initialising) {
            
            int index = hangTimeComboBox.getSelectedIndex();
            
            // Get the previous settings but AND-out the hang time part.
            int previousPinConfig = 0xCF & Integer.parseInt(s.WINKEY_PIN_CONFIG, 16);
            
            int pinConfig = previousPinConfig | (index << 4);
            
            // Overwrite saved settings with new value.
            s.WINKEY_PIN_CONFIG = Integer.toHexString(pinConfig);
            // Send new setting to Winkey.
            wsc.setPinConfig((byte) Integer.parseInt(s.WINKEY_PIN_CONFIG, 16));
            
        }
    }//GEN-LAST:event_hangTimeComboBoxActionPerformed

    private void keyCompSpinnerStateChanged(javax.swing.event.ChangeEvent evt) {//GEN-FIRST:event_keyCompSpinnerStateChanged
        // Convert the displayed value to the code required for Winkey.
        if (!initialising) {
            
            int keyComp = Integer.parseInt(keyCompSpinner.getValue().toString());
            // Overwrite saved settings with new value.
            s.WINKEY_KEY_COMP = Integer.toHexString(keyComp);
            // Send new setting to Winkey.
            wsc.setKeyComp((byte) Integer.parseInt(s.WINKEY_KEY_COMP, 16));
        }
    }//GEN-LAST:event_keyCompSpinnerStateChanged

    private void firstExtnSpinnerStateChanged(javax.swing.event.ChangeEvent evt) {//GEN-FIRST:event_firstExtnSpinnerStateChanged
        // Convert the displayed value to the code required for Winkey.
        if (!initialising) {
            
            int firstExtn = Integer.parseInt(firstExtnSpinner.getValue().toString());
            // Overwrite saved settings with new value.
            s.WINKEY_EXTN1 = Integer.toHexString(firstExtn);
            // Send new setting to Winkey.
            wsc.setFirstExtn((byte) Integer.parseInt(s.WINKEY_EXTN1, 16));
        }
    }//GEN-LAST:event_firstExtnSpinnerStateChanged

    private void restartWinkeyerActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_restartWinkeyerActionPerformed
        // Close & restart.
        wsc.closeWinkey();
        restartWinkey();
    }//GEN-LAST:event_restartWinkeyerActionPerformed

    private void controlsPttRadioButtonActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_controlsPttRadioButtonActionPerformed
        // Update settings.
        s.CWDAEMON_CONTROLS_PTT = "true";
    }//GEN-LAST:event_controlsPttRadioButtonActionPerformed

    private void controlsKeyRadioButtonActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_controlsKeyRadioButtonActionPerformed
        // Update settings.
        s.CWDAEMON_CONTROLS_PTT = "false";
    }//GEN-LAST:event_controlsKeyRadioButtonActionPerformed

    private void formComponentMoved(java.awt.event.ComponentEvent evt) {//GEN-FIRST:event_formComponentMoved
        // Update form's saved X/Y settings upon move.
        s.POSITION_X = Integer.toString(this.getLocation().x);
        s.POSITION_Y = Integer.toString(this.getLocation().y);
    }//GEN-LAST:event_formComponentMoved

    private void debugTextAreaKeyPressed(java.awt.event.KeyEvent evt) {//GEN-FIRST:event_debugTextAreaKeyPressed
        // Send typed characters to winkey (cw keyboard mode).
        char keyPressed = evt.getKeyChar();
        if (keyPressed == ((char) KeyEvent.VK_ESCAPE)) {
            // Abort sending.
            wsc.clearBuffer();
        } else if (keyPressed == ((char) KeyEvent.VK_ENTER)) {
            // The enter char aborts sending.
            // But this isn't the required behaviour, so just trap the event.
        } else {
            wsc.sendText(Character.toString(keyPressed));
        }
    }//GEN-LAST:event_debugTextAreaKeyPressed

    
    
    /**
     * @param args the command line arguments
     */
    public static void main(String args[]) {
        /* Set the Nimbus look and feel */
        //<editor-fold defaultstate="collapsed" desc=" Look and feel setting code (optional) ">
        /* If Nimbus (introduced in Java SE 6) is not available, stay with the default look and feel.
         * For details see http://download.oracle.com/javase/tutorial/uiswing/lookandfeel/plaf.html 
         */
        try {
            for (javax.swing.UIManager.LookAndFeelInfo info : javax.swing.UIManager.getInstalledLookAndFeels()) {
                if ("Nimbus".equals(info.getName())) {
                    javax.swing.UIManager.setLookAndFeel(info.getClassName());
                    break;
                }
            }
        } catch (ClassNotFoundException ex) {
            java.util.logging.Logger.getLogger(WinkeyServerGUI.class.getName()).log(java.util.logging.Level.SEVERE, null, ex);
        } catch (InstantiationException ex) {
            java.util.logging.Logger.getLogger(WinkeyServerGUI.class.getName()).log(java.util.logging.Level.SEVERE, null, ex);
        } catch (IllegalAccessException ex) {
            java.util.logging.Logger.getLogger(WinkeyServerGUI.class.getName()).log(java.util.logging.Level.SEVERE, null, ex);
        } catch (javax.swing.UnsupportedLookAndFeelException ex) {
            java.util.logging.Logger.getLogger(WinkeyServerGUI.class.getName()).log(java.util.logging.Level.SEVERE, null, ex);
        }
        //</editor-fold>

        /* Create and display the form */
        java.awt.EventQueue.invokeLater(new Runnable() {
            public void run() {
                final WinkeyServerGUI serverGUI = new WinkeyServerGUI();
                serverGUI.setVisible(true);
                serverGUI.addWindowListener(new java.awt.event.WindowAdapter() {
                    @Override
                    public void windowClosing(java.awt.event.WindowEvent e) {
                        // Save settings to file.
                        serverGUI.close();
                        Settings.getInstance().writeIni();
                        System.exit(0);
                    }
                });
            }
        });
    }
    // Variables declaration - do not modify//GEN-BEGIN:variables
    private javax.swing.JCheckBox autospaceCheckBox;
    private javax.swing.JPanel basicPanel;
    private javax.swing.JCheckBox contestSpacingCheckBox;
    private javax.swing.JRadioButton controlsKeyRadioButton;
    private javax.swing.JRadioButton controlsPttRadioButton;
    private javax.swing.JPanel cwControlPanel;
    private javax.swing.JTextField cwdaemonPortTextField;
    private javax.swing.JSpinner dahDitRatioSpinner;
    private javax.swing.JTextArea debugTextArea;
    private javax.swing.JSpinner farnsWpmSpinner;
    private javax.swing.JSpinner firstExtnSpinner;
    private javax.swing.JComboBox hangTimeComboBox;
    private javax.swing.JRadioButton hostSetsSpeedButton;
    private javax.swing.JTextField hostWpmTextField;
    private javax.swing.JLabel jLabel1;
    private javax.swing.JLabel jLabel10;
    private javax.swing.JLabel jLabel11;
    private javax.swing.JLabel jLabel12;
    private javax.swing.JLabel jLabel13;
    private javax.swing.JLabel jLabel14;
    private javax.swing.JLabel jLabel15;
    private javax.swing.JLabel jLabel16;
    private javax.swing.JLabel jLabel17;
    private javax.swing.JLabel jLabel2;
    private javax.swing.JLabel jLabel3;
    private javax.swing.JLabel jLabel4;
    private javax.swing.JLabel jLabel5;
    private javax.swing.JLabel jLabel6;
    private javax.swing.JLabel jLabel7;
    private javax.swing.JLabel jLabel8;
    private javax.swing.JLabel jLabel9;
    private javax.swing.JPanel jPanel1;
    private javax.swing.JPanel jPanel2;
    private javax.swing.JScrollPane jScrollPane1;
    private javax.swing.JTabbedPane jTabbedPane1;
    private javax.swing.JSpinner keyCompSpinner;
    private javax.swing.JComboBox keyModeComboBox;
    private javax.swing.JCheckBox keyOut1CheckBox;
    private javax.swing.JCheckBox keyOut2CheckBox;
    private javax.swing.JTextField maxWpmTextField;
    private javax.swing.JTextField minWpmTextField;
    private javax.swing.JCheckBox paddleOnlySidetoneCheckBox;
    private javax.swing.JCheckBox paddleSwapCheckBox;
    private javax.swing.JSpinner paddleSwitchpointSpinner;
    private javax.swing.JCheckBox paddleWatchDogDisableCheckBox;
    private javax.swing.JRadioButton potSetsSpeedButton;
    private javax.swing.JTextField potWpmTextField;
    private javax.swing.ButtonGroup pttButtonGroup;
    private javax.swing.JCheckBox pttEnableCheckBox;
    private javax.swing.JSpinner pttLeadSpinner;
    private javax.swing.JSpinner pttTailSpinner;
    private javax.swing.JButton restartWinkeyer;
    private javax.swing.JComboBox serialPortComboBox;
    private javax.swing.JComboBox sideToneFreqComboBox;
    private javax.swing.JCheckBox sidetoneEnableCheckBox;
    private javax.swing.ButtonGroup speedButtonGroup;
    private javax.swing.JPanel txControlPanel;
    private javax.swing.JSpinner weightingSpinner;
    // End of variables declaration//GEN-END:variables


    //Inner class to listen for Winkeyer status.
    private class WinkeyObserver implements Observer {

        public void update(Observable o, final Object arg) {
            
            if (((ArrayList) arg).get(0).equals("speedPot")) {
                
                if (s.DEBUG.equalsIgnoreCase("true")) {
                debugTextArea.append("Received a speedpot byte: 0x" +
                        ((ArrayList) arg).get(1).toString() + "\n");
                }

                // Need to calculate the speed as an int.
                int potWpm = Integer.parseInt(((ArrayList) arg).get(1).toString(), 16)
                            + Integer.parseInt(s.WINKEY_MIN_WPM, 16);

                potWpmTextField.setText(Integer.toString(potWpm));
                
            } else if (((ArrayList) arg).get(0).equals("status")) {
                
                // There could be more than one status bit set.
                Iterator e = ((ArrayList) arg).iterator();
                while (e.hasNext()) {
                    e.next();   // Dummy read to remove 0th entry.
                    if (s.DEBUG.equalsIgnoreCase("true")) {
                    debugTextArea.append("Received status byte: " +
                        e.next().toString() + "\n");
                    }
                }
                
            } else if (((ArrayList) arg).get(0).equals("serialEcho")) {
                
                serialEchoByte = Integer.parseInt(((ArrayList) arg).get(1).toString(), 16);
                serialEchoPending = true;
                if (s.DEBUG.equalsIgnoreCase("true")) {
                debugTextArea.append("Received serialEcho byte: 0x"
                        + Integer.toHexString(serialEchoByte) + "\n");
                }

            } else if (((ArrayList) arg).get(0).equals("version")) {
                
                version = Integer.parseInt(((ArrayList) arg).get(1).toString(), 16);
                versionPending = true;
                debugTextArea.append("Winkey version: 0x"
                        + Integer.toHexString(version) + "\n");
                
            }
            

        }// update.
        
    }// WinkeyObserver.
    
    
    //Inner class to listen for UDP server events.
    private class UdpServerObserver implements Observer {

        public void update(Observable o, final Object arg) {
            
            if (((ArrayList) arg).get(0).equals("speed")) {
                
                debugTextArea.append(((ArrayList) arg).get(1).toString());
                debugTextArea.append(": requested speed change\n");
                hostWpmTextField.setText(((ArrayList) arg).get(2).toString());
                // Kick-off changes to stored settings & winkey.
                hostWpmTextFieldChanged();
                
            } else if (((ArrayList) arg).get(0).equals("tone")) {
                
                debugTextArea.append(((ArrayList) arg).get(1).toString());
                debugTextArea.append(": requested tone change\n");
                
                int tone = Integer.parseInt(((ArrayList) arg).get(2).toString());
                if (tone == 0) {
                    // Turn off sidetone.
                    sidetoneEnableCheckBox.setSelected(false);
                    // Kick-off changes to stored settings & winkey.
                    sidetoneEnableCheckBoxChanged();
                } else {
                    // Turn on sidetone.
                    sidetoneEnableCheckBox.setSelected(true);
                    sidetoneEnableCheckBoxChanged();
                    
                    // And pick the closest appropriate sidetone frequency.
                    int index = sideToneFreqComboBox.getItemCount() - 1;
                    for (int i = 0; i < sideToneFreqComboBox.getItemCount(); i++) {
                        if (tone <= Integer.parseInt((String) sideToneFreqComboBox.getItemAt(i))) {
                            index = i;
                            break;
                        }
                    }
                    sideToneFreqComboBox.setSelectedIndex(index);
                    // Kick-off changes to stored settings & winkey.
                    sideToneFreqComboBoxChanged();
                }
                
            } else if (((ArrayList) arg).get(0).equals("abort")) {
                
                debugTextArea.append(((ArrayList) arg).get(1).toString());
                debugTextArea.append(": requested message abort\n");
                wsc.clearBuffer();
                
            } else if (((ArrayList) arg).get(0).equals("weight")) {
                
                debugTextArea.append(((ArrayList) arg).get(1).toString());
                debugTextArea.append(": requested weight change\n");
                
                // cwdaemon uses -50 ~ +50 with 0 as no weighting.
                // winkey uses +10 ~ +90 with +50 as no weighting.
                int weight = Integer.parseInt(((ArrayList) arg).get(2).toString()) + 50;
                if (weight < 10) {
                    weight = 10;
                } else if (weight > 90) {
                    weight = 90;
                }
                
                weightingSpinner.setValue(weight);
                // Kick-off changes to stored settings & winkey.
                weightingSpinnerChanged();
                
            } else if (((ArrayList) arg).get(0).equals("ptt")) {
                
                int ptt = Integer.parseInt(((ArrayList) arg).get(2).toString());
                // Send the PTT state to winkey.
                if (s.CWDAEMON_CONTROLS_PTT.equalsIgnoreCase("true")) {
                    wsc.setPtt((byte) ptt);
                } else {
                    wsc.keyImmediate((byte) ptt);
                }
                
            } else if (((ArrayList) arg).get(0).equals("tune")) {
                
                int keyDuration = Integer.parseInt(((ArrayList) arg).get(2).toString());
                wsc.keyBuffered((byte) keyDuration);
                
            } else if (((ArrayList) arg).get(0).equals("sound")) {
                
                int soundState = Integer.parseInt(((ArrayList) arg).get(2).toString());
                if (soundState == 0) {
                    // Turn off sidetone.
                    sidetoneEnableCheckBox.setSelected(false);
                    // Kick-off changes to stored settings & winkey.
                    sidetoneEnableCheckBoxChanged();
                } else {
                    // Turn on sidetone.
                    sidetoneEnableCheckBox.setSelected(true);
                    sidetoneEnableCheckBoxChanged();
                }
                
            } else if (((ArrayList) arg).get(0).equals("message")) {
                
                debugTextArea.append(((ArrayList) arg).get(1).toString());
                debugTextArea.append(": " + ((ArrayList) arg).get(2).toString() + "\n");
                
                // Need to convert cwdaemon style text commands to winkey style.
                String cwText = ((ArrayList) arg).get(2).toString();
                
                // SN, AS & BK prosigns do not map to an existing  winkey prosign.
                // But winkey allows prosigns to be created 'on the fly'. All that's
                // needed is to prepend the two letters with an <ESC> char. This is
                // interpreted as command 1B (merge letters).
                // SN:
                StringBuilder prosign_SN = new StringBuilder();
                prosign_SN.append((char) KeyEvent.VK_ESCAPE);
                prosign_SN.append("SN");                
                // AS:
                StringBuilder prosign_AS = new StringBuilder();
                prosign_AS.append((char) KeyEvent.VK_ESCAPE);
                prosign_AS.append("AS");
                // BK:
                StringBuilder prosign_BK = new StringBuilder();
                prosign_BK.append((char) KeyEvent.VK_ESCAPE);
                prosign_BK.append("BK");
                
                
                // Note: the order of conversion is important, because of prosign code clashes.
                cwText = cwText.replace('~', '|');
                cwText = cwText.replace("!", prosign_SN.toString());
                cwText = cwText.replace("&", prosign_AS.toString());
                cwText = cwText.replace(">", prosign_BK.toString());
                cwText = cwText.replace('<', '>');
                cwText = cwText.replace('*', '<');
                                
                //
                // Look for +/- chars that indicate +/-2wpm speed changes.
                // Do this last because it will split packet into smaller packets.
                //
                // First get the speed we'll start with.
                int startSpeed;
                if (hostSetsSpeedButton.isSelected()) {
                    startSpeed = Integer.parseInt(hostWpmTextField.getText());
                } else {
                    startSpeed = Integer.parseInt(potWpmTextField.getText());
                }
                
                // Now split into tokens using +/- as the delimiter (keep delimiter chars).
                StringTokenizer tokens = new StringTokenizer(cwText, "+-", true);
                int speed = 0;
                while (tokens.hasMoreTokens()) {
                    String token = tokens.nextToken();
                    if (token.equalsIgnoreCase("+")) {
                        // +2wpm.
                        speed = speed + 2;
                    } else if (token.equalsIgnoreCase("-")) {
                        // -2wpm.
                        speed = speed - 2;
                    } else {
                        // Plain text.
                        if (speed != 0) {
                            // Send buffered speed change to winkey.
                            wsc.bufferedSpeedChange((byte) (startSpeed + speed));
                            //System.out.println(startSpeed + speed);
                            // Send text to winkey.
                            wsc.sendText(token);
                        } else {
                            // Send text to winkey.
                            wsc.sendText(token);
                        }
                    }
                }
                // Send cancel buffered speed change to winkey.
                wsc.cancelBufferedSpeedChange();
                
            }
                    

        }// update.

    }// UdpServerObserver.
    
}